#ifndef FPA_H_INCLUDED
#define FPA_H_INCLUDED

typedef unsigned char uchar;

typedef struct
{
  int size;
  int *len;
  void ***data;
} FPA;

void fpa_init(FPA *F, int *len, void ***data, int isize);
void fpa_put(FPA *P, void *ptr);

#endif

#ifndef _djgkjdioweoif_ptrbuf_h_included
#define _djgkjdioweoif_ptrbuf_h_included

typedef struct ptrbuf ptrbuf_t;

#ifndef _PTRBUF_PROTO
#define _PTRBUF_PROTO 
#endif

_PTRBUF_PROTO ptrbuf_t *ptrbuf_new(size_t blksize);
_PTRBUF_PROTO void ptrbuf_free(ptrbuf_t *B, void ***Rd, size_t *Rl);
_PTRBUF_PROTO void ptrbuf_collect(ptrbuf_t *B, void ***Rd, size_t *Rl);
_PTRBUF_PROTO void ptrbuf_put(ptrbuf_t *B, void *d);
_PTRBUF_PROTO void* ptrbuf_get(ptrbuf_t *B);
_PTRBUF_PROTO size_t ptrbuf_count(ptrbuf_t *B);

#endif

#ifndef xalloc_h_included
#define xalloc_h_included
void          *xalloc  (size_t sz);
void          *xallocr (void *ptr, size_t sz);
void          *xallocz (size_t sz);
char          *xstrdup (const char *s);
void           xfree   (void *p);
unsigned char *xstrdupu(const unsigned char *s);
#endif

#ifndef CRITBIT_H_INCLUDED
#define CRITBIT_H_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
  void *root;
} strmap_t;

strmap_t *strmap_new();
void *strmap_find(strmap_t *map, char *key);
void *strmap_insert(strmap_t *map, char *key,void *data);
void *strmap_remove(strmap_t *map, char *key);
void *strmap_destroy(strmap_t *map);

#ifdef __cplusplus
};
#endif

#endif



typedef struct
{
  void* child[2];
  unsigned internal:1, byte:23, mask:8;
} strmap_node_t;


static inline int strmap_dir
    (uint8_t *key,uint32_t len_key,strmap_node_t *node)
{
  uint8_t c;

  if (node->byte<len_key) c= key[node->byte];
                     else c= 0;
  return (1+(node->mask|c))>>8;
}

void* strmap_find(strmap_t *map,char *skey)
{
  uint8_t *key; unsigned int len_key;
  strmap_node_t *node;

  node= map->root;
  if (!node) return 0;

  key= (void*) skey;
  len_key= strlen(skey);

  while(node->internal)
     node= node->child[strmap_dir(key,len_key,node)];

  if (strcmp(skey, node->child[0])) return 0;
                               else return node->child[1];
}


static inline uint32_t strmap_make_mask
    (uint32_t A, uint32_t B)
{
  uint8_t diff;
  diff= A ^ B;
  diff|= diff>>1;
  diff|= diff>>2;
  diff|= diff>>4;
  return (diff&~(diff>>1))^255;
}

static inline strmap_node_t* strmap_leaf_node
    (uint8_t *key, void *data)
{
  strmap_node_t *N;
  N= malloc(sizeof(*N));
  N->internal= 0;
  N->child[0]= key;
  N->child[1]= data;
  return N;
}

static inline strmap_node_t* strmap_internal_node
    (uint32_t byte, uint32_t mask)
{
  strmap_node_t *N;
  N= malloc(sizeof(*N));
  N->internal= 1;
  N->byte= byte;
  N->mask= mask;
  return N;
}

void* strmap_insert(strmap_t* map, char* skey, void *data)
{
  uint8_t *key; uint32_t len_key;
  strmap_node_t *node_elt;

  key= (void*) skey;

  if (!map->root)
  {
     map->root= strmap_leaf_node(key,data);
     return 0;
  }

  len_key= strlen(skey);

  node_elt= map->root;
  while(node_elt->internal)
    node_elt= node_elt->child[strmap_dir(key,len_key,node_elt)];

  uint32_t crit_byte;
  uint8_t  mask;
  uint8_t  *elt;
  int dir_elt;
  strmap_node_t *node_key;

  elt= node_elt->child[0];

  for(crit_byte=0;
      crit_byte<=len_key && elt[crit_byte]==key[crit_byte];
      crit_byte++)
          ;

  if (crit_byte>len_key) return node_elt->child[1];

  mask= strmap_make_mask(elt[crit_byte],key[crit_byte]);

  dir_elt= (1+(mask|elt[crit_byte]))>>8;

  node_key= strmap_internal_node(crit_byte,mask);
  node_key->child[1-dir_elt]= strmap_leaf_node(key,data);

  strmap_node_t *parent;
  strmap_node_t *node;

  parent= 0;
  node= map->root;

  while(node->internal)
  {
     if (node->byte > crit_byte) break;
     if (node->byte==crit_byte && node->mask > mask) break;
     parent= node;
     node= node->child[strmap_dir(key,len_key,node)];
  }

  node_key->child[dir_elt]= node;
  if (parent) { if (parent->child[0]==node) parent->child[0]= node_key;
                                    else    parent->child[1]= node_key; }
  else map->root= node_key;
  return 0;
}

void *strmap_remove(strmap_t* map, char *skey)
{
  uint8_t *key; uint32_t len_key;
  void *retv;

  if (!map->root) return 0;

  key= (void*) skey;
  len_key= strlen(skey);
   
  strmap_node_t *node;
  strmap_node_t *sibling;
  strmap_node_t *parent;
  strmap_node_t *gparent;

  node= map->root;
  gparent= 0;
  sibling= 0;
  parent= 0;

  while(node->internal)
  {
    int dir;
    dir= strmap_dir(key, len_key, node);
    gparent= parent;
    parent= node;
    sibling= node->child[1-dir];
    node= node->child[dir];
  }

  if (strcmp(node->child[0],skey)) return 0;
  retv= node->child[1];
  free(node);

  if (!gparent)
  {
     map->root= sibling;
     if (parent) free(parent);
     return retv;
  }

  if (gparent->child[0]==parent)
    gparent->child[0]= sibling;
  else
    gparent->child[1]= sibling;

  free(parent);

  return retv;
}

void *strmap_destroy(strmap_t *map)
{
  strmap_node_t *node;
  strmap_node_t *parent;
  strmap_node_t *gparent;
  strmap_node_t *sibling;
  void *retv;

  if (!map->root) return 0;

  parent= 0;
  gparent= 0;
  sibling= 0;
  node= map->root;
  while(node->internal)
  {
    gparent= parent;
    parent= node;
    sibling= node->child[1];
    node= node->child[0];
  }
  retv= node->child[1];
  free(node);
  if (!gparent) 
  {
    map->root= sibling;
    if (parent) free(parent);
    return retv;
  }
  if (gparent->child[0]==parent)
    gparent->child[0]= sibling;
  else
    gparent->child[1]= sibling;

  free(parent);

  return retv;
}

strmap_t  *strmap_new()
{
  strmap_t *s;
  s= malloc(sizeof(*s));
  s->root= 0;
  return s;
}



static void *memdie(void *R)
{
  if (R) return R;
  fprintf(stderr,"out of memory\n");
  exit(1);
}
void   *xalloc  (size_t sz) { return memdie(malloc(sz)); }
void   *xallocr (void *ptr, size_t sz) { return memdie(realloc(ptr, sz)); }
void   *xallocz (size_t sz) { return memdie(calloc(1,sz)); }
char   *xstrdup (const char *s) { return memdie(strdup(s)); }
unsigned char *xstrdupu(const unsigned char *s) 
         { return memdie(strdup((const char*)s)); }
       
void    xfree   (void *p) { if (p) free(p); }


struct pbufnode { 
  struct pbufnode *next;
  size_t len, start;
  void* data[];
};

struct ptrbuf {
  struct pbufnode *first, *last;
  size_t blksize;
};


_PTRBUF_PROTO ptrbuf_t *ptrbuf_new(size_t blksize)
{
  ptrbuf_t *B;
  B= malloc(sizeof(*B));
  B->first= B->last= NULL;
  B->blksize= blksize;
  return B;
}


_PTRBUF_PROTO void ptrbuf_free(ptrbuf_t *B, void ***Rd, size_t *Rl)
{
  struct pbufnode *P,*N;
  if (Rd || Rl) ptrbuf_collect(B, Rd, Rl);
  for(P=B->first;P;P=N)
  {
    N= P->next;
    free(P);
  }
  free(B);
}

_PTRBUF_PROTO void ptrbuf_collect(ptrbuf_t *B, void ***Rd, size_t *Rl)
{
  size_t S;
  struct pbufnode *P;
  void **R;

  if (!Rd && !Rl) return ;

  S= 0;
  for(P=B->first;P;P=P->next) S+= P->len;

  if (S==0) { if (Rd) *Rd= NULL; if (Rl) *Rl= 0; return ; }
  if (!Rd) { if (Rl) *Rl= S; return ; }

  R= malloc(S*sizeof(void*));
  S= 0;
  for(P=B->first;P;P=P->next)
  {
    memcpy(R+S, P->data+P->start, P->len*sizeof(void*));
    S+= P->len;
  }

  *Rd= R;
  if (Rl) *Rl= S;
}

_PTRBUF_PROTO void ptrbuf_put(ptrbuf_t *B, void *d)
{
  struct pbufnode *P;

  P= B->last;
  if (!P || (P->len+P->start)==B->blksize)
  {
      P= malloc(sizeof(*P)+B->blksize*sizeof(void*));
      P->len= P->start= 0;
      P->next= NULL;
      if (!B->last) { B->last= B->first= P; }
      else { B->last->next= P; B->last= P; }
 }
 P->data[P->len+P->start]= d;
 P->len++;
}

_PTRBUF_PROTO void* ptrbuf_get(ptrbuf_t *B)
{
  struct pbufnode *P;
  void *R;

  P= B->first;
  if (!P) return NULL;
  R= P->data[P->start];
  P->start++;
  P->len--;
  if (P->len==0)
  {
    B->first= P->next;
    if (!B->first) B->last= NULL;
    free(P);
  }
  return R;
}

_PTRBUF_PROTO size_t ptrbuf_count(ptrbuf_t *B)
{
  size_t R;
  ptrbuf_collect(B, NULL, &R);
  return R;
}


void fpa_init(FPA *F, int *len, void ***data, int isize)
{
  F->len= len;
  F->data= data;
  F->size= isize;
  *F->data= xalloc(sizeof(void*)*isize);
  F->len[0]= 0;
}

void fpa_put(FPA *P, void *ptr)
{
  int L;
  L= P->len[0];
  if (L==P->size)
  {
    P->size += P->size/2; 
    P->data[0]= xallocr(P->data[0], sizeof(void*)*P->size);
  } 
  P->data[0][L++]= ptr;
  P->len[0]= L;
}


int ustrcmp(uchar *a, char *b) { return strcmp((char*) a, b); }




static void bsort_ptr(void **a, int n, int (*compare)(void *, void *))
{
  int i;
  void *t;
  for(;n>1;n--)
    for(i=0;i<n-1;i++)
      if ((*compare)(a[i],a[i+1])>0)
        { t= a[i]; a[i]= a[i+1]; a[i+1]= t; }
}
