

      // values for the coord field.
enum { wc_left, wc_top, wc_width, wc_height, wc_right, wc_bottom };

// coordinate name to coordinate code
// converts a string to one of the enums above
int cn2cc(unsigned char *cname, unsigned char *pos);
int ce2cc(expr_t *E)
{
  return cn2cc(E->data, E->pos);
}

char *cc2cn(int cc)
{
  switch(cc)
  {
  case wc_left: return "left";
  case wc_top:  return "top";
  case wc_width: return "width";
  case wc_height: return "height";
  case wc_right: return "right";
  case wc_bottom: return "bottom";
  }
  return "unkn-coord-nam";
}

//
// returns 1 if the expression is of the form .xx
// and xx is a coordinate name
// coordinate code is stored in *cc if it's so.
int ceB2cc(expr_t *E, int *cc)
{
  int C;
  if (!expr_match(E, node_self_field, NULL, 0)) return 0;
  C= cn2cc(E->data,NULL);
  if (C==-1) return 0;
  if (cc) *cc= C; 
  return 1;
}

//
// same as above, but now the expression is of the
// form foo.xx where foo is a kidget name
//
int ceA2cc(expr_t *E, uchar *wname, int *cc)
{
  int C;
  if (!expr_match(E, node_field, NULL, 2)) return 0;
  if (!expr_match(E->sub[0], node_ident, wname, 0)) return 0;
  if (!expr_match(E->sub[1], node_ident, NULL, 0)) return 0;
  C= cn2cc(E->sub[1]->data, NULL);
  if (C==-1) return 0;
  if (cc) *cc= C;
  return 1;
}

typedef struct  value value_t;

//
// these are access types.
// a value can be stored as a local or global array
// as well as in the coordinates array
// it can also be a constant zero value.
//  
enum { at_zero, at_coord, at_local, at_global };

//
// these are value types.
// if a value is a coordinate, it can be computed
// by alternate means if an expression for it is
// not given
//
enum { v_var, v_coord };

struct value
{
  int type;

  int Atype;     // access type
  int Aarg[2];   // access arg

  enum { vs_unknown, vs_visited, vs_known } state; 

  unsigned char *pos;

  //
  // These are valid when type==v_coord
  //
  unsigned char *winame;     int cc;

  char *key;
  expr_t *expr;
};

// values are stored in a tree. a key is associated with each value.
// this is done by encoding the type field, variable name field,
// kidget id and coordinate id. Since constants are always known,
// we don't need to look them up.

static char *vkey_global(char *name)
{
  char buf[200];
  snprintf(buf,sizeof(buf)-1,"G:%s", name);
  buf[sizeof(buf)-1]= 0;
  return xstrdup(buf);
}

static char *vkey_local(char *name)
{
  char buf[200];
  snprintf(buf,sizeof(buf)-1,"L:%s", name);
  buf[sizeof(buf)-1]= 0;
  return xstrdup(buf);
}

static char *vkey_wcoord(uchar *name, int cooridx)
{
  char buf[200];
  snprintf(buf,sizeof(buf)-1,"C%d:W%s", cooridx, name);
  buf[sizeof(buf)-1]= 0;
  return xstrdup(buf);
}

static char *vkey_scoord(int cooridx)
{
  char buf[200];
  snprintf(buf,sizeof(buf)-1,"S%d", cooridx);
  buf[sizeof(buf)-1]= 0;
  return xstrdup(buf);
}


int cn2cc(unsigned char *cname, unsigned char *pos)
{
  static const struct sipair names[]=
  {
    { "left", wc_left },
    { "l"  , wc_left },
    { "x"  , wc_left },
    { "top", wc_top  },
    { "t"  , wc_top  },
    { "y"  , wc_top  },
    { "width", wc_width },
    { "w", wc_width },
    { "height", wc_height },
    { "h", wc_height },
    { "right", wc_right },
    { "r",  wc_right },
    { "bottom", wc_bottom },
    { "b", wc_bottom },
    { NULL, 0 },
  };
  int i;
  for(i=0;names[i].sval;i++)
     if (!strcmp((char*) cname, names[i].sval))
        return names[i].ival;
  if (pos)
    irror(pos, "unknown coordinate name %s\n", cname);
  return -1;
}

static void make_coord_accessor(value_t *V, int widx, int cc)
{
  int op;
  V->Atype= at_coord;
  V->Aarg[0]= widx;
  switch(cc)
  {
  case wc_left:   op= GUII_WLAY_X; break;
  case wc_top:    op= GUII_WLAY_Y; break;
  case wc_width:  op= GUII_WLAY_W; break;
  case wc_height: op= GUII_WLAY_H; break;
  }
  V->Aarg[1]= op;
}

static void make_local_accessor(value_t *V, int idx)
{
  V->Atype= at_local;
  V->Aarg[0]= idx;
}

void make_global_accessor(value_t *V, int idx)
{
  V->Atype= at_global;
  V->Aarg[0]= idx;
}

static void make_const0_accessor(value_t *V)
{
  V->Atype= at_zero;
}

static void access_value(value_t *V, int rW, struct byco *out)
{
  switch(V->Atype)
  {
  case at_zero:
    if (rW)  byco_op(out, GUII_DROP);
       else  byco_int(out, 0);
    return;
  case at_local:
    byco_int(out,V->Aarg[0]);
    byco_op(out, GUII_LOCAL);
    break;
  case at_global:
    byco_int(out,V->Aarg[0]);
    byco_op(out, GUII_GLOBAL);
    break;
  case at_coord:
    byco_int(out, V->Aarg[0]);
    byco_op(out, V->Aarg[1]);
    break;
  }
  byco_op(out, rW ? GUII_STOREINT : GUII_LOADINT);
}

value_t *value_new(int type, char *key, uchar *pos)
{
  value_t *V;
  V= xallocz(sizeof(*V));
  V->type= type;
  V->key= key;
  V->pos= pos;
  return V;
}
