
void init_file_option_i(file_option_t *O, char *name, int value)
{
  O->name= name;
  O->value.type= v_int;
  O->value.intv= value;
}

void init_file_option_s(file_option_t *O, char *name, char* value)
{
  O->name= name;
  O->value.type= v_str;
  O->value.strv= value;
}

int record_file_option(input_t *ipt, char *name, expr_t *value)
{
  static const int offsets[]= {
      offsetof(input_t,o_window_width),
      offsetof(input_t,o_window_height),
      offsetof(input_t,o_window_title),
      offsetof(input_t,o_android_activity),
      offsetof(input_t,o_freetype),
      offsetof(input_t,o_png_read),
      offsetof(input_t,o_png_write),
      -1 };
  file_option_t *O;
  int i;

  if (strncmp(name,"opt_", 4)) return 0;
  
  for(i=0;offsets[i]>=0;i++)
  {
    O= (void*) ipt+offsets[i];
    if (!strcmp(name, O->name))
       break;
  }
  if (offsets[i]<0)
     irror(value->pos, "unknown option '%s'\n", name);

  if (O->pos)
     irror2(O->pos, value->pos, "option '%s' was set twice\n", name);

  O->pos= value->pos;
  switch(O->value.type)
  {
  case v_int:
    if (value->type!=node_number)
       irror(O->pos, "option '%s' needs an integer constant\n", name);
    O->value.intv= atoi((char*) value->data);
    break;
  case v_str:
    if (value->type!=node_string)
       irror(O->pos, "option '%s' needs a string constant\n", name);
    O->value.strv= xstrdup((char*) value->data);
    break;
  default:
    irror(O->pos,"internal: unhandled option type %d\n", O->value.type);
  }
  return 1;
}

void record_file_options(input_t *ipt)
{
  int i;
  expr_t *E,*L,*R; 

  init_file_option_i(&ipt->o_window_width, "opt_window_width", 800);
  init_file_option_i(&ipt->o_window_height, "opt_window_height", 600);
  init_file_option_s(&ipt->o_window_title, "opt_window_title", "GFX Window");
  init_file_option_s(&ipt->o_android_activity, "opt_android_activity", "Android.Activity");
  init_file_option_i(&ipt->o_png_read, "opt_png_read", 0);
  init_file_option_i(&ipt->o_png_write, "opt_png_read", 0);
  init_file_option_i(&ipt->o_freetype, "opt_freetype", 0);

  for(i=0;i<ipt->n_assign;i++) 
  { 
    E= ipt->assign[i];
    if (!expr_match(E, node_assign, NULL, 2)) continue; // bad assignment
    L= E->sub[0];
    R= E->sub[1];
                        // must be an assignment to an identifier
    if (!expr_match(L, node_ident, NULL, 0)) continue;
    if (!record_file_option(ipt, (char*) L->data, R)) continue;  
    ipt->assign[i--]= ipt->assign[--ipt->n_assign];
  }
}
