

static void eval_circular_eq(value_t *V)
{
  irror(V->pos, "circular definition detected\n");
}

static void eval_value
   (econtext_t *C, value_t *V,struct byco *code,int getval);

static void ecompute_altn_param
   (econtext_t *C, value_t *R, int coord, struct byco *code)
{
  value_t *V;
  V= con_lookup_field(C, R->winame, coord, R->pos);
  if (!V->expr)
     irror(R->pos, "undefined coordinate %s.%s\n", R->winame, cc2cn(coord));
  eval_value(C, V, code, 1);
}

static void ecompute_altn(econtext_t *C, value_t *V, struct byco *code)
{
  int c1, c2, op;
  switch(V->cc)
  {
  case wc_left:   c1= wc_right;  op= GUII_SUB; c2= wc_width;   break;
  case wc_right:  c1= wc_left;   op= GUII_ADD; c2= wc_width;   break;
  case wc_width:  c1= wc_right;  op= GUII_SUB; c2= wc_left;    break;
  case wc_top:    c1= wc_bottom; op= GUII_SUB; c2= wc_height;  break;
  case wc_bottom: c1= wc_top;    op= GUII_ADD; c2= wc_height;  break;
  case wc_height: c1= wc_bottom; op= GUII_SUB; c2= wc_top;     break;
  default:
    irror(V->pos, "internal: invalid coordinate\n");
  }

  ecompute_altn_param(C, V, c1, code);
  ecompute_altn_param(C, V, c2, code);
  byco_op( code, op );
}

static void eval_invalid_expr(expr_t *E)
{
   unsigned char *pos;
   pos= E->pos;
   switch(E->type)
   {
   case node_self_field: 
     irror(pos, "internal: self_field not rewritten\n");
   case node_list:
   case node_string: 
     irror(pos, "non-numeric value in expression\n");
   case node_assign:
     irror(pos, "assignment does not produce a value\n");
   case node_comma:
     irror(pos, "internal: unprocessed comma operator\n");
   case node_call:
     irror(pos, "internal: unimplemented operator: 'call'\n");
   }
}

static void eval_expr(econtext_t *C, expr_t *E, struct byco *code);

static void eval_value
   (econtext_t *C, value_t *V, struct byco *code, int getval)
{
  if (V->state==vs_visited) eval_circular_eq(V);
  if (V->state==vs_known) 
  {
    if (getval) access_value(V, 0, code); 
    return ;
  }

  V->state= vs_visited;

  if (V->expr)
  {
     eval_expr(C, V->expr, code);
  }
  else
  {
     if (V->type==v_coord)
        ecompute_altn(C, V, code); 
     else
        irror(V->pos, "unassigned variable\n");
  }
  if (getval) byco_op(code, GUII_DUP);
  access_value(V, 1, code);
  V->state= vs_known;
}

static void eval_expr(econtext_t *C, expr_t *E, struct byco *code)
{
  int inst;
  switch(E->type)
  {
    case node_list:  case node_string: 
   case node_assign:     case node_comma: case node_call:
     eval_invalid_expr(E);
     break;
   case node_number:
     if (strchr((char*) E->data,'.'))
       byco_float(code, atof((char*) E->data));
     else
       byco_int(code, atoi((char*) E->data));
     break;
   case node_self_field:
   case node_ident: 
   case node_field:
      eval_value(C, con_lookup(C, E), code, 1);
      break;
   case node_sub: inst= GUII_SUB; goto binop;
   case node_add: inst= GUII_ADD; goto binop;
   case node_div: inst= GUII_DIV; goto binop;
   case node_mod: inst= GUII_MOD; goto binop;
   case node_mul: inst= GUII_MUL; goto binop;
   case node_negate: 
     byco_int(code, 0);
     eval_expr(C, E->sub[0], code);
     byco_op(code, GUII_SUB);
     break;
  }
  return ;
binop:
  eval_expr(C, E->sub[0], code);
  eval_expr(C, E->sub[1], code);
  byco_op(code, inst);
}



void record_assignments_in_context
    (ptrbuf_t *P, econtext_t *C,
     int n_expr, expr_t **expr, struct byco *code)
{
  int i;
  expr_t *E;
  value_t *V;
  int free_p;

  if (code && !P) { P= ptrbuf_new(500); free_p= 1; }
             else free_p= 0;

  for(i=0;i<n_expr;i++)
  {
    E= expr[i];
    V= con_lookup(C, E->sub[0]);
    if (V->expr) 
       irror2(V->pos, E->pos, "Multiple assignments to same value\n");
    if (V->state==vs_known)
       irror(E->pos, "Assignment to system variable\n");
    V->expr= E->sub[1];
    if (P) ptrbuf_put(P, V);
  }

  if (code && P)
    while((V=ptrbuf_get(P)))
      eval_value(C, V, code, 0);

  if (free_p)
     ptrbuf_free(P, NULL, NULL);
}
 
