

typedef struct dis
{
  unsigned int intv;
  int count;
  FILE *out;
  void (*func)(struct dis *, unsigned char B);
  int p_nl;
} dis_t;

void d_nl(dis_t *D,int nl)
{
  if (D->p_nl) fprintf(D->out,"\n");
  D->p_nl= nl;
}

void d_out(dis_t *D, int nl, const char *fmt, ...)
{
  d_nl(D,nl);
  va_list ap;
  va_start(ap,fmt);
  vfprintf(D->out, fmt, ap);
  va_end(ap);
}

void d_byte(dis_t *D, unsigned char B) { (*D->func)(D,B); }

void d_initial(dis_t *D, unsigned char B);

void d_reset(dis_t *D)
{
  D->func= d_initial;
  D->count= 0;
}

void d_count(dis_t *D, int nl, const char *fmt, ...)
{
  va_list ap;
  D->count--;
  if (D->count) return ;
  if (fmt)
  {
    d_nl(D, nl);
    va_start(ap, fmt);
    vfprintf(D->out, fmt, ap);
    va_end(ap);
  }
  d_reset(D);
}

void d_skip(dis_t *D, unsigned char B)
{
  d_count(D, 0, NULL);
}

void d_int(dis_t *D, unsigned char B)
{
  D->intv= (D->intv << 8) | B;
  d_count(D, 1, "(int) 0x%x ", D->intv);
}

void d_str(dis_t *D, unsigned char B)
{
  if (B) fwrite(&B, 1,1, D->out);
  d_count(D, 1, "> ");
}

void d_flo(dis_t *D, unsigned char B)
{
  d_out(D, 0, "%02x ", B);
  d_count(D, 1, " "); 
}

void d_tab(dis_t *D, int nl, char *name, uchar B, char **tab)
{
  if (tab[B]) d_out(D, nl, "%s %s ", name, tab[B]);
         else d_out(D, nl, "%s %02x ", name, B);
  d_reset(D);
}

void d_section(dis_t *D, uchar B)
{
  static char *secnam[256]=
  {
  [GUISEC_SCREENDIM]= "SCREENDIM",
  [GUISEC_CONSTRUCTOR]= "CONSTRUCTOR",
  [GUISEC_WIDGETLAYOUT]= "WIDGETLAYOUT",
  };
  d_tab(D, 1, "section", B, secnam);
}

void d_setg(dis_t *D, uchar B)
{
  static char *gnam[256]=
  {
    [GUIG_WINDOW_WIDTH]= "WINDOW_WIDTH",
    [GUIG_WINDOW_HEIGHT]= "WINDOW_HEIGHT",
    [GUIG_WINDOW_TITLE]= "WINDOW_TITLE",
    [GUIG_NFONT]= "NFONT",
    [GUIG_NFONTFILE]= "NFONTFILE",
    [GUIG_NICON]= "NICON",
    [GUIG_NSCREEN]= "NSCREEN",
    [GUIG_FIRST_WIDGET]= "FIRST_WIDGET",
    [GUIG_LAST_WIDGET]= "LAST_WIDGET",
    [GUIG_NGLOBAL]= "NGLOBAL",
  };
  d_tab(D, 1, "setg", B, gnam);
}

void d_initial(dis_t *D, unsigned char B)
{
static char *inam[256]=
{
[GUII_GLOBALS_SECTION] "GLOBALS_SECTION",
[GUII_C4]= "C4",
[GUII_C3]= "C3",
[GUII_C3]= "C3",
[GUII_C2]= "C2",
[GUII_C1]= "C1",
[GUII_ADD]= "ADD",
[GUII_SUB]= "SUB",
[GUII_MUL]= "MUL",
[GUII_DIV]= "DIV",
[GUII_MOD]= "MOD",
[GUII_STOREINT]= "STOREINT",
[GUII_LOADINT]= "LOADINT",
[GUII_LOCAL]= "LOCAL",
[GUII_WLAY_X]= "WLAY_X",
[GUII_WLAY_Y]= "WLAY_Y",
[GUII_WLAY_W]= "WLAY_W",
[GUII_WLAY_H]= "WLAY_H",
[GUII_CF]= "CF",
[GUII_GLOBAL]= "GLOBAL",
[GUII_STR]= "STR",
[GUII_CONSTRUCT]= "CONSTRUCT",
[GUII_INTARRAY]= "INTARRAY",
[GUII_IPARAM]= "IPARAM",
[GUII_DONE]= "DONE",
[GUII_NOP]= "NOP",
[GUII_PPARAM]= "PPARAM",
[GUII_SETG] = "SETG",
[GUII_FONTFILE]= "FONTFILE",
[GUII_FONT]= "FONT",
[GUII_ICON]= "ICON",
[GUII_ICONIMAGE]= "ICONIMAGE",
[GUII_BLOB]= "BLOB",
[GUII_SECTION]= "SECTION",
[GUII_SCRINFO]= "SCRINFO",
[GUII_SUBLAYOUT]= "SUBLAYOUT",
[GUII_DUP] = "DUP",
[GUII_XADD] = "XADD",
[GUII_LOADX] = "LOADX",
};
  switch(B)
  {
  case GUII_C4: D->intv= 0; D->count= 4; D->func= d_int; break; 
  case GUII_C3: D->intv= 0; D->count= 3; D->func= d_int; break;
  case GUII_C2: D->intv= 0; D->count= 2; D->func= d_int; break;
  case GUII_C1: D->intv= 0; D->count= 1; D->func= d_int; break;
  case GUII_STR: D->count= D->intv; D->func= d_str;
                 d_out(D, 0, "(str) <"); break;
  case GUII_CF: D->count= 4; D->func= d_flo;
                d_out(D, 0, "(flo)"); break; 
  case GUII_SECTION: D->func= d_section; break;
  case GUII_SETG: D->func= d_setg; break;
  case GUII_BLOB: D->count= D->intv; D->func= d_skip;
                  d_out(D, 1, "(blob) %d ", D->count); break;
  case GUII_STOREINT:
  case GUII_LOADINT:
  case GUII_LOCAL:
  case GUII_WLAY_X:
  case GUII_WLAY_Y:
  case GUII_WLAY_W:
  case GUII_WLAY_H:
  case GUII_GLOBAL:
  case GUII_LOADX:
  case GUII_XADD:
       D->p_nl= 0;
  default:
    if (B&0x80)
    {
      D->intv= B ^ 0x80;
      d_out(D, 1, "(int) 0x%x ", D->intv);
    }
    else
    {
      if (inam[B]) d_out(D, 1, "%s ", inam[B]);
              else d_out(D, 1, "unknown %02x ", B);
    }
  }
}

dis_t *d_init(FILE *out)
{
  dis_t *D;
  D= xallocz(sizeof(*D));
  D->out= out;
  D->func= d_initial;
  return D;
}


void d_done(dis_t *D)
{
  if (D->p_nl) fprintf(D->out,"\n");
  free(D);
}


