
#define DEFAULT_FONTFILE0_NAME ":builtin-normal:"
#define DEFAULT_FONTFILE1_NAME ":builtin-bold:"
#define DEFAULT_FONT0_SIZE     14
#define DEFAULT_FONT1_SIZE     14
#define DEFAULT_FONT2_SIZE     20


void make_fontfile_array(input_t *ipt, sipair_t **ff, int *nff)
{
  simap_t *files;
  int i;
  char *path;

  files= simap_new();

  simap_getp(files,DEFAULT_FONTFILE0_NAME); 
  simap_getp(files,DEFAULT_FONTFILE1_NAME);
  for(i=0;i<ipt->n_font;i++)
  {
    path= path_sibling(get_pos_ifile(ipt->font[i]->pos), ipt->font[i]->file);
    ipt->font[i]->fino= simap_getp(files, path);
    free(path);
  }
  simap_free(files, ff, nff, 'I');
}

void free_fontfile_array(sipair_t *ff, int nff)
{
  int i;
  for(i=0;i<nff;i++) free(ff[i].sval);
  free(ff);
}

void gen_font_files(input_t *ipt)
{
  int nff;
  sipair_t *ff;
  int i;
  int file_size;

  make_fontfile_array(ipt, &ff, &nff);

  byco_setg_i(ipt->code, GUIG_NFONTFILE, nff);

  for(i=2;i<nff;i++)
  {
    file_size= byco_file_blob(ipt->code, ff[i].sval);
    byco_int(ipt->code, file_size);
    byco_int(ipt->code, i); 
    byco_op(ipt->code, GUII_FONTFILE);
  }

  free_fontfile_array(ff, nff);
}

static void make_default_fonts(kontdef_t **font, int nfont)
{
  int i;

  static const struct
  {
    int fino, size;
  } defont[]= {
     { 0,  DEFAULT_FONT0_SIZE },
     { 1,  DEFAULT_FONT1_SIZE },
     { 1,  DEFAULT_FONT2_SIZE },
  };

  for(i=0;i<3;i++)
  {
    if (font[i]) continue;
    font[i]= xallocz(sizeof(kontdef_t));
    font[i]->fino= defont[i].fino;
    font[i]->size= defont[i].size;
  }
}

void get_fonts(input_t *ipt, kontdef_t ***def, int *ndef)
{
  int nfont;
  kontdef_t **font;
  int num;
  int i;

  nfont= 2;
  for(i=0;i<ipt->n_font;i++)
    nfont= imax2(ipt->font[i]->number-1,nfont);
  nfont++;

  font= xallocz(nfont * sizeof(kontdef_t*));

  for(i=0;i<ipt->n_font;i++)
  {
    num= ipt->font[i]->number-1;
    if (num<0 || num>99)
       irror(ipt->font[i]->pos, "invalid font number %d\n", num+1);
    if (font[num])
       irror2(font[num]->pos, ipt->font[i]->pos,
                "font %d defined multiple times\n", num+1);
    font[num]= ipt->font[i];
  }

  make_default_fonts(font, nfont);


  *def= font;
  *ndef= nfont;
}

void gen_fonts(input_t *ipt)
{
  int i;
  int nfont;
  kontdef_t **font;

  gen_font_files(ipt);
  get_fonts(ipt, &font, &nfont);

  byco_setg_i(ipt->code, GUIG_NFONT, nfont);

  for(i=0;i<nfont;i++) 
  {
    if (!font[i]) continue;
    byco_int(ipt->code, font[i]->size);
    byco_int(ipt->code, font[i]->fino);
    byco_int(ipt->code, i);
    byco_op(ipt->code, GUII_FONT);
  }

  free(font);
}

