expr_t *get_attrib(expr_t **A, int nA, char *name)
{
  int i;
  expr_t *E, *R;
  for(i=0;i<nA;i++)
  {
    E= A[i];
    if (!expr_match(E, node_assign, NULL, 2)) continue;
    R= E->sub[1];
    E= E->sub[0];
    if (!expr_match(E, node_self_field, (uchar*) name, 0)) continue;
    return R;
  }
  return NULL;
}

expr_t *expr_new_kidget_coor(expr_t *W, uchar *pos, char *fname)
{
  return expr_new_subs( node_field, pos, 2, 
                        W,
                        expr_new_leaf(node_ident, (uchar*) fname, pos));
}

void check_align_opts(int opt, uchar *pos)
{
  const int STACK= (ALIGNOPT_LR| ALIGNOPT_RL| ALIGNOPT_TD| ALIGNOPT_BU);
  const int EQHOR= (ALIGNOPT_EL | ALIGNOPT_ER);
  const int EQVER= (ALIGNOPT_ET | ALIGNOPT_EB);

  switch(opt & STACK)
  {
            case 0: break;
  case ALIGNOPT_LR:
  case ALIGNOPT_RL: if (opt&EQHOR) goto invalid; break;
  case ALIGNOPT_TD:
  case ALIGNOPT_BU: if (opt&EQVER) goto invalid; break;
           default: goto invalid;
  }

  if ((opt&EQHOR)==EQHOR && (opt&ALIGNOPT_EW)) goto invalid;
  if ((opt&EQVER)==EQVER && (opt&ALIGNOPT_EH)) goto invalid;
  return ;

invalid:
  irror(pos, "invalid set of stack options\n");
}

void make_assign(ptrbuf_t *out, uchar *pos, expr_t *L, expr_t *R)
{
  ptrbuf_put(out, expr_new_subs( node_assign, pos, 2, L, R));
}

void stack_kidgets(align_t *A, ptrbuf_t *out)
{
  int i;
  expr_t *W, *N, *lexpr, *rexpr, *list, *first, *pad;
  char *asgn, *val; int op;
  const int STACK= (ALIGNOPT_LR| ALIGNOPT_RL| ALIGNOPT_TD| ALIGNOPT_BU);

  switch(A->options & STACK)
  {
  case 0: return ;
  case ALIGNOPT_LR: asgn= "left"; val= "right"; op= node_add; break;
  case ALIGNOPT_RL: asgn= "right"; val= "left"; op= node_sub; break;
  case ALIGNOPT_TD: asgn= "top"; val= "bottom"; op= node_add; break;
  case ALIGNOPT_BU: asgn= "bottom"; val= "top"; op= node_sub; break;
  }

  list= A->kidget_list;

  first= get_attrib(A->assign, A->n_assign, "first");
  if (first)
  {
     lexpr= expr_new_kidget_coor(list->sub[0], A->pos, asgn);
     make_assign(out, A->pos, lexpr, first);
  }

  pad= get_attrib(A->assign, A->n_assign, "pad");

  W= list->sub[0];
  for(i=1;i<list->nsub;i++)
  {
    N= list->sub[i];
    
    lexpr= expr_new_kidget_coor( N, A->pos, asgn );    
    rexpr= expr_new_kidget_coor( W, A->pos, val );
    if (pad)
       rexpr= expr_new_subs( op, A->pos, 2, rexpr, pad);
    make_assign(out, A->pos, lexpr, rexpr);
    W= N;
  }
}

void equalize_kidget_coords(ptrbuf_t *out, expr_t *L, expr_t *R, char *fname)
{
  expr_t *lexpr, *rexpr;
  lexpr= expr_new_kidget_coor( L, L->pos, fname );    
  rexpr= expr_new_kidget_coor( R, R->pos, fname );    
  make_assign(out, L->pos, lexpr, rexpr);
}

void equalize_coords(align_t *A, ptrbuf_t *out)
{
  int i, opt;
  expr_t *W, *N, *list;
  const int EQ= ALIGNOPT_EL | ALIGNOPT_ER |
                ALIGNOPT_ET | ALIGNOPT_EB |
                ALIGNOPT_EW | ALIGNOPT_EH;

  opt= A->options;
  if (!(opt & EQ)) return ;

  list= A->kidget_list;
  W= list->sub[0];
  
  for(i=1;i<list->nsub;i++)
  {
    N= list->sub[i];
    if (opt&ALIGNOPT_EW) equalize_kidget_coords(out, N, W, "width");
    if (opt&ALIGNOPT_EH) equalize_kidget_coords(out, N, W, "height");
    if (opt&ALIGNOPT_ET) equalize_kidget_coords(out, N, W, "top");
    if (opt&ALIGNOPT_EB) equalize_kidget_coords(out, N, W, "bottom");
    if (opt&ALIGNOPT_EL) equalize_kidget_coords(out, N, W, "left");
    if (opt&ALIGNOPT_ER) equalize_kidget_coords(out, N, W, "right");
  }
}

void process_align(align_t *A, ptrbuf_t *out)
{
  check_align_opts(A->options, A->pos);
  if (A->kidget_list->nsub<2)
     irror(A->pos, "single element list is invalid\n");
  stack_kidgets(A, out);
  equalize_coords(A, out);
}

void expand_align(kreen_t *S)
{
  int i;
  ptrbuf_t *P;

  if (!S->n_align) goto no_align;
 
  P= ptrbuf_new(50);
  for(i=0;i<S->n_align;i++) process_align(S->align[i], P);
  kreen_merge_assign(S, P);

no_align:
  for(i=0;i<S->n_allsub;i++) expand_align(S->allsub[i]);
}
