

/**
  the kreen object has the following fields related to the computation
  of kreen geometry:

  int n_assign;     expr_t **assign;
  int n_eslay;      expr_t **eslay;
  struct byco *slay; // kreen layout

  after running the functions here, assignments in 'assign' should
  be transferred to 'eslay'.

  the system will then compile expressions in 'eslay' into 'slay'.

  changing the geometry inside a subkreen doesn't make sense, since
  it's already determined by the containing kidget. for those assignments,
  we give a warning and continue.

  when transferring assignments, we look at expressions of the form

   kreen.w = <expr>
   kreen.x=  <expr>
      or
   .top= <expr>

  etc. These are all converted to 'node_self_field' type expressions.

 **/


int move_kreen_coord(expr_t *E, ptrbuf_t *P)
{
  expr_t *L;
  uchar *pos;
  int cc;

  if (!(L=is_assign(E))) return 0;
  pos= E->pos;

  if (ceA2cc(L, (uchar*) "screen",&cc))
  {
    E= expr_new_assign(pos,
                 expr_new_self_field(pos, (uchar*) cc2cn(cc)),
                 E->sub[1]);
  }
  else 
  {
    if (!ceB2cc(L,NULL))
       return 0;
  }

  ptrbuf_put(P, E);
  return 1;
}

void extract_kreen_coords(kreen_t *S, int is_sub)
{
  ptrbuf_t *P;
  int i;

  P= ptrbuf_new(10);
  for(i=0;i<S->n_assign;i++)
    if (move_kreen_coord(S->assign[i], P))
       S->assign[i--]= S->assign[--S->n_assign];
  exprbuf_free(P, &S->n_eslay, &S->eslay);

  if (is_sub)
  {
    for(i=0;i<S->n_eslay;i++)
      iwarn(S->eslay[i]->pos, "subkreen geometry computations are ignored\n");
    S->n_eslay= 0;
    S->eslay= NULL;
  }
  else
  {
    for(i=0;i<S->n_allsub;i++)
       extract_kreen_coords(S->allsub[i], 1);
  }
}
