
/**
  coordinate equations for kidgets should be extracted to the kreen level
  and separated into 'ewlay' array. this is because kidget layout is
  computed separately from everything else.

  first step is to extract assignments within kidget definitions outside.
  for example

   button foo "OK"   .x= bar.x;

  should be converted into

   button foo "OK";
   foo.x= bar.x;

  while doing this, right side of the equation needs to be processed also.
  any self-field nodes should be converted into field nodes with the kidget
  name prefixed.
 **/

 /** if E is a self_field, it's replaced with a field, target being
     object name as an identifier expression, ready to be put into
     the replacement **/
expr_t *sf_replace_f(expr_t *E, expr_t *target)
{
  int i;
  uchar *C;

  if ((C=is_self_field_x(E)))
    return expr_new_field_ed(E->pos, target, C);

  for(i=0;i<E->nsub;i++)
    E->sub[i]= sf_replace_f(E->sub[i], target);

  return E;
}

expr_t *replident(expr_t *E, char *fin, uchar *rep)
{
  int i;
  if (!rep) return E;
  if (is_ident(E, (uchar*) fin)) return expr_new_ident(rep,E->pos);
  for(i=0;i<E->nsub;i++)
    E->sub[i]= replident(E->sub[i], fin, rep);
  return E;
}

/***
  extracts coordinate equations from kidget object.
  both   kidget.x = expr
  and          .x = expr
  style of equations are extracted.
 ***/
void ewc_from_kidget(kidget_t *W, ptrbuf_t *P, uchar *srep)
{
  int i;
  expr_t *L, *E;
  uchar *wname;
 
  wname= (uchar*) W->ident;

  for(i=0;i<W->n_assign;i++)
  {
    E= W->assign[i];
    if (!(L=is_assign(E))) continue;
    if (!ceB2cc(L,NULL) && !ceA2cc(L, wname,NULL)) continue;
    ptrbuf_put(P, replident(
                     sf_replace_f(E, expr_new_ident(wname, E->pos)),
                     "screen", srep) );
    W->assign[i--]= W->assign[--W->n_assign];
  }
}

void ewc_from_kreen(kreen_t *S, ptrbuf_t *P, uchar *srep)
{
  int i;
  expr_t *E, *L;
  for(i=0;i<S->n_assign;i++)
  {
    E= S->assign[i];
    if ((L=is_assign(E)) && (ceA2cc(L, NULL, NULL)||is_ident(L,NULL)))
    {
      ptrbuf_put(P, replident(E,"screen", srep));
      S->assign[i--]= S->assign[--S->n_assign];
    }
  }
}

uchar *parent_name(kreen_t *S)
{
  return S->subname==-1? NULL : S->name;
}

void extract_kidget_coords(kreen_t *S)
{
  ptrbuf_t *P;
  int i, j;
  kreen_t *K;
  P= ptrbuf_new(1000);

  ewc_from_kreen(S, P, NULL);
  for(i=0;i<S->n_kidget;i++)
     ewc_from_kidget(S->kidget[i], P, NULL);

  for(i=0;i<S->n_allsub;i++)
  {
    K= S->allsub[i];
    ewc_from_kreen(K, P, K->name);
    for(j=0;j<K->n_kidget;j++)
       ewc_from_kidget(K->kidget[j], P, K->name);
  }
  exprbuf_free(P, &S->n_ewlay, &S->ewlay);
}

/***
  for sub-screens, the process is more involved.

  for a widget W in subscreen U, which is part of a panel P,
  all references to 
    screen.w and screen.h should be replaced with  P.w and P.h
 ***/
