

static char *wa_find_p
  (kidget_t *W, char *name, int type, char *msg, uchar **pos)
{
  int i;
  expr_t *value, *E;

  value= NULL;
  for(i=0;!value && i<W->n_assign;i++)
  {
    E= W->assign[i];
    if (!is_assign(E)) continue;
    if (is_self_field(E->sub[0], (uchar*) name))
      value= E->sub[1];
    else if (is_field(E->sub[0], (uchar*) W->ident, (uchar*) name))
      value= E->sub[1];
  }
  if (!value)
     return NULL;

  if (pos) *pos= value->pos;
  if (type!=value->type)
     irror(value->pos, "expected %s for field '%s'\n", msg, name);
  return (char*) value->data;
}

static char *wa_find(kidget_t *W, char *name, int type, char *msg)
{
  return wa_find_p(W,name,type, msg, NULL);
}

bcarg_t wa_int(kidget_t *W, char *name, int dv)
{
  char *data;
  data= wa_find(W, name, node_number, "number");
  if (data) 
     dv= strtol(data, NULL, 0);
  return bcINT(dv);
}

bcarg_t wa_float(kidget_t *W, char *name, float dv)
{
  char *data;
  data= wa_find(W, name, node_number, "number");
  if (data) 
     dv= strtof(data, NULL);
  return bcFLO(dv);
}

bcarg_t wa_str(kidget_t *W, char *name, char* dv)
{
  char *data;
  data= wa_find(W, name, node_string, "string");
  if (data) 
    dv= data;
  return bcSTR(dv); 
}

bcarg_t wa_text(kidget_t *W, char *dv)
{
  return wa_str(W, "text", dv); 
}

bcarg_t wa_enum(kidget_t *W, char *name, const sipair_t *values, int dv)
{
  char *data;
  uchar *pos;
  data= wa_find_p(W, name, node_ident, "identifier", &pos);
  if (data)
  {
    dv= sipair_lookupnc_a(values, data);
    if (dv<0) 
       irror(pos, "unknown enumeration value\n");
  }
  return bcINT(dv);
}

bcarg_t wa_bool(kidget_t *W, char *name, int dv)
{
  const static sipair_t values[]=
  {
   { "yes", 1 }, { "on" , 1 }, { "enabled", 1 }, { "true" , 1 },
   { "no", 0 }, { "off", 0 }, { "disabled", 0 }, { "false", 0 }, 
   { NULL, -1 }
  };
  return wa_enum(W, name, values, dv);
}

bcarg_t wa_justify(kidget_t *W, int dv)
{
  const static sipair_t values[]=
  {
    { "center", 'C' }, { "left",   'L' }, { "right",  'R' },
    { NULL, -1 }
  };
  return wa_enum(W, "justify", values, dv);
}

bcarg_t wa_iconpos(kidget_t *W, int dv)
{
  const static sipair_t values[]=
  {
    { "top", 'T' }, { "left",   'L' }, { "right",  'R' }, {"bottom", 'B'},
    { NULL, -1 }
  };
  return wa_enum(W, "justify", values, dv);
}

bcarg_t wa_icon(input_t *ipt, kidget_t *W, int dv)
{
  int i;
  char *data;
  uchar *pos;

  data= wa_find_p(W, "icon", node_ident, "icon", &pos);
  if (!data)
     return bcINT(dv);
  for(i=0;i<ipt->n_icon;i++)
    if (!strcmp((char*) ipt->icon[i]->ident, data))
       return bcINT(i);
  irror(pos, "unknown icon '%s'\n", data);
  return bcINT(dv);
}


bcarg_t wa_orient(kidget_t *W, int dv)
{
  const static sipair_t values[]=
  {
    { "horizontal", 'H' },
    { "vertical",   'V' },
    { NULL, -1 }
  };
  return wa_enum(W, "orientation", values, dv);
}

bcarg_t wa_kidget_id(kreen_t *S, kidget_t *W, char *name, int dv)
{
  char *wname;
  uchar *pos;
  int i;
  wname= wa_find_p(W, name, node_ident, "kidget_id", &pos);
  if (!wname) return bcINT(dv);
  for(i=0;i<S->n_allkidget;i++)
    if (!strcmp((char*) S->allkidget[i]->ident, wname))
       return bcINT(S->allkidget[i]->global_name);
  irror(pos, "unknown kidget '%s'\n", wname);
  return bcINT(dv);
}

