
struct program_options
{
  char *dc, *dh, *cc, *ch, *lc, *lh, *t;
  int f, pr, pw, d, v, T, g, h, help;
} prg;

static const struct
{
  char *name;
  int  type;
  void *ptr;
  char *description;
} option_defs[]=
{
  { "dc",  'F', &prg.dc, ".c data"},
  { "dh",  'F', &prg.dh, ".h data"},
  { "cc",  'F', &prg.cc, "combined library + data source" },
  { "ch",  'F', &prg.ch, "combined library + data header" },
  { "lc",  'F', &prg.lc, "library source"},
  { "lh",  'F', &prg.lh, "library header"},
  { "f",   'B', &prg.f,  "enable freetype functions"},
  { "pr",  'B', &prg.pr, "enable lodepng decoding functions"},
  { "pw",  'B', &prg.pw, "enable lodepng encoding functions"},
  { "d",   'B', &prg.d,  "disassemble generated code"},
  { "v",   'B', &prg.v,  "print version info"},
  { "h",   'B', &prg.h,  "program options help"},
  { "help", 'B', &prg.help, "detailed usage text"},
  { "g",   'B', &prg.g,  "generate sample input" },
  { "t", 'S', &prg.t,  "test given screen"},
  { "T",  'B', &prg.T, "test all screens"},
  { NULL, 0, NULL, NULL}
};

void print_options_help(char *pname, int quit, int err)
{
  int i;
  FILE *f;
  f= err ? stderr : stdout;
  fprintf(f, "Usage: %s input_file <options>\n", pname);
  fprintf(f, "Options:\n");
  for(i=0;option_defs[i].name;i++)
  {
    fprintf(f,"  -%s ", option_defs[i].name);
    switch(option_defs[i].type)
    {
    case 'S': fprintf(f,"<screen> "); break;
    case 'F': fprintf(f,"<file> "); break;
    }
    fprintf(f,"%s\n", option_defs[i].description);
  }
  if (quit) exit(1);
}

static int find_opt(char *opname)
{
  int j;
  for(j=0;option_defs[j].name;j++)
    if (!strcmp(option_defs[j].name, opname))
       return j;
  die("unknown option '%s'\n", opname);
  return -1;
}

static void shift_arg(int *argc, char **argv, int s)
{
  int i;
  for(i=s;i<*argc-1;i++)
    argv[i]= argv[i+1];
  (*argc)--;
}

void get_optarg(int oindex, int desc, int *argc, char **argv)
{
  switch(option_defs[desc].type)
  {
  case 'B': *(int*) option_defs[desc].ptr= 1; return ;
  case 'S': 
  case 'F':
    break;
  }
  if (oindex>=*argc || argv[oindex][0]=='-')
     die("option '%s' needs an argument\n", option_defs[desc].name);
  *(char**) option_defs[desc].ptr= argv[oindex];
  shift_arg(argc, argv, oindex);
}

void parse_program_options(int *argc, char **argv)
{
  int i, d;
  for(i=1;i<*argc;i++)
  {
    if (argv[i][0]!='-') continue;
    d= find_opt(argv[i]+1);
    shift_arg(argc, argv, i);
    get_optarg(i, d, argc, argv);
    i--;
  }
}
