
#define OUT_FILE   1
#define OUT_MULTI  2

typedef struct out
{
  int type;
  FILE *f;
  int n_sub;
  struct out *sub[20];

  int last;
} out_t;

out_t *out_new(int type)
{
  out_t *O;
  O= xallocz(sizeof(*O));
  O->type= type;
  return O;
}

out_t *out_file(char *fn)
{
  out_t *O;
  if (!fn) return NULL;
  O= out_new(OUT_FILE);
  O->f= fopen(fn, "wb");
  if (!O->f) dies("open(%s): ", fn);
  return O;
}

out_t *out_multi()
{
  return out_new(OUT_MULTI);
}

void ovprintf(out_t *O, const char *fmt, va_list ap)
{
  int i;
  switch(O->type)
  {
  case OUT_FILE: vfprintf(O->f, fmt, ap); break;
  case OUT_MULTI:
    for(i=0;i<O->n_sub;i++)
      ovprintf(O->sub[i], fmt, ap);
    break;
  }
}

void oprintf(out_t *O, const char *fmt, ...)
{
  va_list ap;
  va_start(ap,fmt);
  ovprintf(O, fmt, ap);
  va_end(ap);
}

void owrite(out_t *O, void *buf, int sz)
{
  int i;
  switch(O->type)
  {
  case OUT_FILE: fwrite(buf, 1, sz, O->f); break; 
  case OUT_MULTI:
    for(i=0;i<O->n_sub;i++)
      owrite(O->sub[i], buf, sz);
    break;
  }
}


void out_close(out_t *O)
{
  int i;
  if (!O) return ;
  switch(O->type)
  {
  case OUT_FILE: fclose(O->f); break; 
  case OUT_MULTI:
    for(i=0;i<O->n_sub;i++)
      out_close(O->sub[i]);
    break;
  }
  free(O);
}

int out_choose(out_t **R, int n, ...)
{
  out_t *O;
  out_t *V;
  int i;
  va_list ap;
  O= out_multi();
  va_start(ap,n);
  for(i=0;i<n;i++)
  {
    V= va_arg(ap, out_t*);
    if (V) O->sub[O->n_sub++]= V;
  }
  va_end(ap);
  switch(O->n_sub)
  {
  case 0: free(O); return 0;
  case 1: *R= O->sub[0]; free(O); return 1;
  }
  *R= O;
  return 1;
}

void oenum_begin(out_t *O, char *name)
{
  oprintf(O, "enum %s {\n", name);
  O->last= -1;
}

void oenum_end(out_t *O)
{
  oprintf(O,"};\n");
}

void oenum(out_t *O, char *name, int value)
{
  if (value==O->last+1)
     oprintf(O, "  %s,\n", name);
  else
     oprintf(O, "  %s= %d,\n", name, value);
  O->last= value;
}

