

kreen_t *ikreen(char *name, ptrbuf_t *kidgets)
{
  kreen_t *S;
  size_t sz;
  S= xallocz(sizeof(*S));
  S->name= (uchar*) name;
  ptrbuf_free(kidgets, (void***) &S->kidget, &sz);
  S->n_kidget= sz;
  return S;
}

kidget_t *ikidget(ptrbuf_t *kidgets, char *name, int klass, ptrbuf_t *asgn)
{
  kidget_t *W;
  size_t sz;
  W= xallocz(sizeof(*W));
  W->ident= name;
  W->klass= klass;
  ptrbuf_free(asgn, (void***) &W->assign, &sz);
  W->n_assign= sz;
  ptrbuf_put(kidgets, W);
  return W;
}

void iattrib(ptrbuf_t *asgn, char *name, expr_t *V)
{
  ptrbuf_put(asgn, expr_new_assign(NULL, 
                       expr_new_self_field(NULL, (uchar*) name),
                       V));
}
expr_t* idiv(expr_t *L, expr_t *R)
{
  return expr_new_binop(node_div, NULL, L, R); 
}
expr_t* iadd(expr_t *L, expr_t *R)
{
  return expr_new_binop(node_add, NULL, L, R); 
}
expr_t* isub(expr_t *L, expr_t *R)
{
  return expr_new_binop(node_sub, NULL, L, R); 
}
expr_t* imul(expr_t *L, expr_t *R)
{
  return expr_new_binop(node_mul, NULL, L, R); 
}
expr_t *inumber(char *data)
{
  return expr_new_leaf(node_number, (uchar*) data, NULL);
}
expr_t* istring(char *data)
{
  return expr_new_leaf(node_string, (uchar*) data, NULL);
}
expr_t* ifield(char *obj, char *field)
{
  return expr_new_field(NULL,
     expr_new_ident((uchar*) obj, NULL),
     expr_new_ident((uchar*) field, NULL));
}
expr_t *ivar(char *ident)
{
  return expr_new_ident((uchar*) ident, NULL);
}
expr_t *ifont1h(char *factor)
{
  return imul( ivar("font1h"), inumber(factor));
}
expr_t *iselff(char *name)
{
  return expr_new_self_field(NULL, (uchar*) name);
}
void icenterx(ptrbuf_t *abuf)
{
  iattrib(abuf, "x", idiv( isub(ifield("screen","w"),
                                iselff("w")),
                           inumber("2")));
}

void imake_test_screen(input_t *ipt)
{
  kreen_t *S;
  ptrbuf_t *wbuf;
  ptrbuf_t *abuf;

  wbuf= ptrbuf_new(50);

  abuf= ptrbuf_new(10);
  iattrib(abuf, "text", istring("List of Screens"));
  iattrib(abuf, "font", inumber("3"));
  iattrib(abuf, "justify", ivar("Center"));
  iattrib(abuf, "x", inumber("0"));
  iattrib(abuf, "y", inumber("0"));
  iattrib(abuf, "h", ifont1h("1.3"));
  iattrib(abuf, "w", ifield("screen", "w"));
  ikidget(wbuf, "_tLABEL", GFXCLASS_LABEL, abuf);

  abuf= ptrbuf_new(10);
  iattrib(abuf, "text", istring("Run"));
  icenterx(abuf); 
  iattrib(abuf, "b", ifield("screen", "h"));
  iattrib(abuf, "h", ifont1h("1.3"));
  iattrib(abuf, "w", ifont1h("8"));
  ikidget(wbuf, "_tRUN", GFXCLASS_BUTTON, abuf);

  abuf= ptrbuf_new(10);
  iattrib(abuf, "x", inumber("0"));
  iattrib(abuf, "y", ifield("_tLABEL", "b"));
  iattrib(abuf, "r", ifield("screen", "w"));
  iattrib(abuf, "b", ifield("_tRUN", "t"));
  ikidget(wbuf, "_tLIST", GFXCLASS_LIST, abuf);

  S= ikreen("_test", wbuf);

  ipt->kreen= xallocr(ipt->kreen, (ipt->n_kreen+1)*sizeof(ipt->kreen[0]));
  ipt->kreen[ipt->n_kreen++]= S;
}

kreen_t* test_find_kreen(input_t *ipt, char *sname)
{
  int i;
  for(i=0;i<ipt->n_kreen;i++)
    if (!ustrcmp(ipt->kreen[i]->name, sname)) 
       return ipt->kreen[i]; 
  die("unknown screen '%s'\n", sname);
  return NULL;
}

void test_populate_list(int wino, input_t *ipt)
{
  int i;
  //
  // the last one is not inserted because that's the _test screen :)
  //
  for(i=0;i<ipt->n_kreen-1;i++)
    gfx_insert(wino, -1, (char*) ipt->kreen[i]->name, NULL, NULL);
}

void test_screen(int number)
{
  gx.screen(number);
  while(1)
  {
    gfx_event_t e;
    if (!gfx_event(&e)) continue;
    if (e.type==GFXEV_QUIT) break;
  }
  gx.leave();
}

int test_find_kidget(kreen_t *S, char *name)
{
  int i;
  for(i=0;i<S->n_kidget;i++)
    if (!strcmp(S->kidget[i]->ident,name)) 
       return S->kidget[i]->global_name;
  die("internal error: can't find widget '%s'\n", name);
  return -1;
}

// input has already been injected with the test screen
// we just need to show it.
void test_input(input_t *ipt, char *sname)
{
  unsigned char *gcode;
  size_t glen;
  kreen_t *test, *scr;
  int wList, wRun;

  scr= sname ? test_find_kreen(ipt, sname) : NULL;
  test= test_find_kreen(ipt, "_test");

  byco_collect(ipt->code,&gcode,&glen);

  gfx_init_x11(NULL, gcode, glen);
  gx.screen(test->number);

  wList= test_find_kidget(test, "_tLIST");
  test_populate_list(wList, ipt);

  wRun= test_find_kidget(test, "_tRUN");

  if (scr) test_screen(scr->number);

  while(1)
  {
    gfx_event_t e;
    if (!gfx_event(&e)) continue;
    if (e.type==GFXEV_QUIT) break;
    if (e.type==wRun)
    {
      int sel;
      sel= gfx_get_selected(wList);
      if (sel!=-1) test_screen(sel);
    }
  }
  gfx_close();
}
