#include <stdint.h>
#include <dirent.h>
#include <stdio.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdlib.h>
#include "gfx.h"

void panel_demo()
{
  int q;
  gfx_event_t e;

  gx.screen(sPANEL);
  gfx_set_state(wpSel0,1);
  q= 0; while(!q) if (gfx_event(&e)) switch(e.type)
  {
  case GFXEV_QUIT: q=1; break;
  case wpSel0: if (e.key) {
                 gfx_set_state(wpSel1,0);
                 gfx_set_state(wpSel2,0);
                 gfx_set_selected(wpPanel,0); } break;
  case wpSel1: if (e.key) {
                 gfx_set_state(wpSel0,0);
                 gfx_set_state(wpSel2,0);
                 gfx_set_selected(wpPanel,1); } break;
  case wpSel2: if (e.key) {
                 gfx_set_state(wpSel0,0);
                 gfx_set_state(wpSel1,0);
                 gfx_set_selected(wpPanel,2); } break;
  }
  gx.leave();
}


void canvas_demo()
{
  int q;
  gfx_event_t e;
  int selecting;
  int x1,y1,x2,y2;
  int x,y,w,h;
  
  q= 0;
  selecting= 0;
  gx.screen(sCANVAS);
  gfx_logview_printf(wcLog, "This is a %s viewer\n", "log");

  gfx_canvas_setsize(wcCanvas, 400,400);
  gfx_canvas_fill_rectangle(wcCanvas, 0xFFffff, 0,0, 400, 400);
  gfx_canvas_draw_rectangle(wcCanvas, 0xFF0000, 20, 20, 300, 300);
  gfx_canvas_draw_rectangle(wcCanvas, 0x000000, 30, 40, 50, 50);
  gfx_canvas_print(wcCanvas,1,0xff00ff,50,80, "Hello, there!");

  while(!q) if (gfx_event(&e)) switch(e.type)
  {
  case GFXEV_QUIT: q=1; break;
  case GFXEV(wcCanvas,BUTTON_PRESS):
    gfx_logview_printf(wcLog,"canvas click: %d (%d,%d)\n", e.button,e.x,e.y);
    break;
  case GFXEV(wcCanvas,BUTTON_RELEASE):
    gfx_logview_printf(wcLog,"canvas release: %d (%d,%d)\n",e.button,e.x,e.y);
    if (e.button!=1) break;
    if (!selecting)
    {
      gfx_canvas_recsel_begin(wcCanvas, &e);
      selecting= 1;
      break;
    }
    gfx_canvas_recsel_end(wcCanvas, &x1, &y1, &x2, &y2);
    x= (x1<x2?x1:x2);
    y= (y1<y2?y1:y2);
    w= (x1+x2)-2*x+1;
    h= (y1+y2)-2*y+1;
    gfx_canvas_draw_rectangle(wcCanvas, 0, x,y,w,h);
    selecting= 0;
    break;
  }
  gx.leave();
}

int mwi;
void msgset(int wno) { mwi= wno; }
void msg(const char *fmt, ...)
{
  char buf[500];
  va_list ap;
  va_start(ap,fmt);
  vsnprintf(buf, sizeof(buf)-1, fmt, ap);
  va_end(ap);
  buf[sizeof(buf)-1]= 0;
  gfx_set_text(mwi, buf);
}

typedef struct
{
  unsigned int *data;
  unsigned w, h;
  char *name;
} icon_t;
int n_icons;
#define SZ_ICONS 200
icon_t icons[SZ_ICONS];

void mk_icon(char *name)
{
  char buf[200];
  unsigned R;
  icon_t *I;
  snprintf(buf,sizeof(buf)-1,"data/icons/%s",name);
  buf[sizeof(buf)-1]= 0;
  if (n_icons==SZ_ICONS) return ;
  I= icons+n_icons;
  if ((R=lodepng_decode32_file((void*)&I->data, &I->w, &I->h, buf)))
    fprintf(stderr,"can not decode %s: %s\n", buf, lodepng_error_text(R));
  I->name= strdup(name);
  n_icons++;
}

int is_png(char *name)
{
  int L;
  L= strlen(name);
  if (L<=4) return 0;
  name= name + L - 4;
  return !strcasecmp(name,".png");
}

void make_icons()
{
  DIR *dir;
  struct dirent *de;
  char *name;
  if (icons[0].w) return ;
  dir= opendir("data/icons");
  if (!dir)
  {
    gfx_ask("Error","!OK", "Can not access data/icons");
    return ;
  }
  while((de=readdir(dir)))
  {
    name= de->d_name;
    if (is_png(name))
      mk_icon(name);
  }
  closedir(dir);
}

void list_demo()
{
  gfx_event_t e;
  int q;
  int i;
  q= 0;
  msgset(wlStatus);
  gx.screen(sLISTS);

  gfx_dirlist_set_directory(wlDir,getenv("HOME"));
  gfx_filelist_set_directory(wlFile, "/");

  make_icons();
  for(i=0;i<n_icons;i++)
    gfx_iconlist_insert(wlIcon,-1, icons[i].name, icons[i].w, icons[i].h,
                                  icons[i].data, 0, NULL);

  while(!q)
  {
    if (!gfx_event(&e)) continue;
    switch(e.type)
    {
    case GFXEV_QUIT: q=1; break;
    case wlDir: msg("Selected dir: %s", e.string); break;
    case wlFile: msg("Selected file: %s", e.string); break;
    }
  }
  gx.leave();
}

void simple_demo()
{
  gfx_event_t e;
  int q;
  char *str;
  int pv;
  q= 0;
  pv= 0;
  msgset(wsStatus);
  gx.screen(sSIMPLE);

  gfx_insert(wsDrop, -1, "First Entry", NULL,NULL);
  gfx_insert(wsDrop, -1, "Second Entry", NULL,NULL);
  gfx_insert(wsDrop, -1, "Third Entry", NULL,NULL);
  gfx_insert(wsDrop, -1, "Fourth Entry", NULL,NULL);
  gfx_set_selected(wsDrop, 0);

  gfx_progressbar_total(wsProgress, 100);

  gfx_insert(wsList, -1, "This is a list", NULL, NULL);
  gfx_insert(wsList, -1, "Normally, it can", NULL, NULL);
  gfx_insert(wsList, -1, "Contain icons, but", NULL, NULL);
  gfx_insert(wsList, -1, "I'm too lazy to", NULL, NULL);
  gfx_insert(wsList, -1, "Find some ", NULL, NULL);

  while(!q)
  {
    if (!gfx_event(&e)) continue;
    switch(e.type)
    {
    case GFXEV_QUIT: q=1; break;
    case wsButton: msg("Button clicked, making progress.");
                   pv= (pv+1)%100;
                   gfx_progressbar_set(wsProgress, pv);
                   break;
    case wsRad1: msg("Radio:1"); break;
    case wsRad2: msg("Radio:2"); break;
    case wsRad3: msg("Radio:3"); break;
    case wsRad4: msg("Radio:4"); break;
    case wsAsk: msg("Dialog result: %d",
          gfx_ask("Confirmation Dialog",
              "!OK!Cancel!Third Button!Fourth Button",
              "This is a message or confirmation dialog, intended\n"
              "for %s messages. It can display up to %d buttons.\n"
              "\nThe text you see here is specified using a format string\n"
              "and the associated arguments in the same way it's done for\n"
              "the printf function.", "short", 8));
          break;
    case wsLoad:
        str= gfx_select_file(0,"/home/", NULL);
        msg("File selection result: %s", str?str:"(NULL)");
        break;
    case wsSave:
        str= gfx_select_file(1,"/home/", NULL);
        msg("File selection result: %s", str?str:"(NULL)");
        break;
    case wsSliderV: gfx_slider_set(wsSliderH,e.flo0); break;
    case wsSliderH: msg("slider: %f",e.flo0); break;
    case wsToggle: msg("Toggled %s", e.key?"on":"off"); break;
    case wsCheck: msg("Check %s", e.key?"on":"off"); break;
    case wsDrop: msg("Dropdown: %d", e.key);  break;
    }
  }
  gx.leave();
}


int main()
{
  gfx_event_t e;
  int q;
  int sel;
  gfx_init(NULL);
  gx.screen(sMAIN);
  q= 0;
  gfx_insert(wmList, -1, "Simple Widgets", NULL, NULL);
  gfx_insert(wmList, -1, "Some Lists", NULL, NULL);
  gfx_insert(wmList, -1, "Canvas, Pixgrid & Log", NULL, NULL);
  gfx_insert(wmList, -1, "Panels ", NULL, NULL);
  while(!q)
  {
    if (!gfx_event(&e)) continue;
    switch(e.type)
    {
    case GFXEV_QUIT: q=1; break;
    case wmRun:
      sel= gfx_get_selected(wmList);
      switch(sel)
      {
      case 0: simple_demo(); break;
      case 1: list_demo(); break;
      case 2: canvas_demo(); break;
      case 3: panel_demo(); break;
      }
    }
  }
  gx.leave();
  gfx_close();
  return 0;
}

