

/***********************************************************************
 *                                                                     *
 *                              UTILITIES                              *
 *                                                                     *
 ***********************************************************************/

#define _MPRINTF_PROTO static

#ifndef _mprintf_h_included
#define _mprintf_h_included

#ifndef _MPRINTF_PROTO 
#define _MPRINTF_PROTO 
#endif

_MPRINTF_PROTO size_t
vgprintf    (void (*func)(void *,const char *, size_t), void *arg,
             char const *fmt, va_list ap);


_MPRINTF_PROTO char *vmprintf(const char *fmt, va_list ap);

#endif
/*-
 * Copyright (c) 1986, 1988, 1991, 1993
 *	The Regents of the University of California.  All rights reserved.
 * (c) UNIX System Laboratories, Inc.
 * All or some portions of this file are derived from material licensed
 * to the University of California by American Telephone and Telegraph
 * Co. or Unix System Laboratories, Inc. and are reproduced herein with
 * the permission of UNIX System Laboratories, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)subr_prf.c	8.3 (Berkeley) 1/21/94
 */

#define NBBY    8               /* number of bits in a byte */
static char const hex2ascii_data[] = "0123456789abcdefghijklmnopqrstuvwxyz";
#define hex2ascii(hex)  (hex2ascii_data[hex])

/* Max number conversion buffer length: a u_quad_t in base 2, plus NUL byte. */
#define MAXNBUF	(sizeof(intmax_t) * NBBY + 1)

/*
 * Put a NUL-terminated ASCII number (base <= 36) in a buffer in reverse
 * order; return an optional length and a pointer to the last character
 * written in the buffer (i.e., the first character of the string).
 * The buffer pointed to by `nbuf' must have length >= MAXNBUF.
 */
static char *
ksprintn(char *nbuf, uintmax_t num, int base, int *lenp, int upper)
{
	char *p, c;

	p = nbuf;
	*p = '\0';
	do {
		c = hex2ascii(num % base);
		*++p = upper ? toupper(c) : c;
	} while (num /= base);
	if (lenp)
		*lenp = p - nbuf;
	return (p);
}

/*
 * Scaled down version of printf(3).
 *
 * Two additional formats:
 *
 * The format %b is supported to decode error registers.
 * Its usage is:
 *
 *	printf("reg=%b\n", regval, "*");
 *
 * where  is the output base expressed as a control character, e.g.
 * \10 gives octal; \20 gives hex.  Each arg is a sequence of characters,
 * the first of which gives the bit number to be inspected (origin 1), and
 * the next characters (up to a control character, i.e. a character <= 32),
 * give the name of the register.  Thus:
 *
 *	kvprintf("reg=%b\n", 3, "\10\2BITTWO\1BITONE\n");
 *
 * would produce output:
 *
 *	reg=3
 *
 * XXX:  %D  -- Hexdump, takes pointer and separator string:
 *		("%6D", ptr, ":")   -> XX:XX:XX:XX:XX:XX
 *		("%*D", len, ptr, " " -> XX XX XX XX ...
 */
_MPRINTF_PROTO size_t
vgprintf(void (*func)(void *,const char *, size_t), void *arg,
         char const *fmt, va_list ap)
{
#define PCHAR(c) do {char cc=(c); (*func)(arg,&cc,1); retval++;} while(0)
	char nbuf[MAXNBUF];
	const char *p, *percent, *q;
	u_char *up;
	int ch, n;
	uintmax_t num;
	int base, lflag, qflag, tmp, width, ladjust, sharpflag, neg, sign, dot;
	int cflag, hflag, jflag, tflag, zflag;
	int dwidth, upper;
	char padc;
	int stop = 0, retval = 0;
        int radix= 10;

	num = 0;

	if (fmt == NULL)
		fmt = "(fmt null)\n";

	if (radix < 2 || radix > 36)
		radix = 10;

	for (;;) {
		padc = ' ';
		width = 0;
		while ((ch = (u_char)*fmt++) != '%' || stop) {
			if (ch == '\0')
				return (retval);
			PCHAR(ch);
		}
		percent = fmt - 1;
		qflag = 0; lflag = 0; ladjust = 0; sharpflag = 0; neg = 0;
		sign = 0; dot = 0; dwidth = 0; upper = 0;
		cflag = 0; hflag = 0; jflag = 0; tflag = 0; zflag = 0;
reswitch:	switch (ch = (u_char)*fmt++) {
		case '.':
			dot = 1;
			goto reswitch;
		case '#':
			sharpflag = 1;
			goto reswitch;
		case '+':
			sign = 1;
			goto reswitch;
		case '-':
			ladjust = 1;
			goto reswitch;
		case '%':
			PCHAR(ch);
			break;
		case '*':
			if (!dot) {
				width = va_arg(ap, int);
				if (width < 0) {
					ladjust = !ladjust;
					width = -width;
				}
			} else {
				dwidth = va_arg(ap, int);
			}
			goto reswitch;
		case '0':
			if (!dot) {
				padc = '0';
				goto reswitch;
			}
		case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
				for (n = 0;; ++fmt) {
					n = n * 10 + ch - '0';
					ch = *fmt;
					if (ch < '0' || ch > '9')
						break;
				}
			if (dot)
				dwidth = n;
			else
				width = n;
			goto reswitch;
		case 'c':
			PCHAR(va_arg(ap, int));
			break;
		case 'D':
			up = va_arg(ap, u_char *);
			p = va_arg(ap, char *);
			if (!width)
				width = 16;
			while(width--) {
				PCHAR(hex2ascii(*up >> 4));
				PCHAR(hex2ascii(*up & 0x0f));
				up++;
				if (width)
					for (q=p;*q;q++)
						PCHAR(*q);
			}
			break;
		case 'd':
		case 'i':
			base = 10;
			sign = 1;
			goto handle_sign;
		case 'h':
			if (hflag) {
				hflag = 0;
				cflag = 1;
			} else
				hflag = 1;
			goto reswitch;
		case 'j':
			jflag = 1;
			goto reswitch;
		case 'l':
			if (lflag) {
				lflag = 0;
				qflag = 1;
			} else
				lflag = 1;
			goto reswitch;
		case 'n':
			if (jflag)
				*(va_arg(ap, intmax_t *)) = retval;
			else if (qflag)
				*(va_arg(ap, quad_t *)) = retval;
			else if (lflag)
				*(va_arg(ap, long *)) = retval;
			else if (zflag)
				*(va_arg(ap, size_t *)) = retval;
			else if (hflag)
				*(va_arg(ap, short *)) = retval;
			else if (cflag)
				*(va_arg(ap, char *)) = retval;
			else
				*(va_arg(ap, int *)) = retval;
			break;
		case 'o':
			base = 8;
			goto handle_nosign;
		case 'p':
			base = 16;
			sharpflag = (width == 0);
			sign = 0;
			num = (uintptr_t)va_arg(ap, void *);
			goto number;
		case 'q':
			qflag = 1;
			goto reswitch;
		case 'r':
			base = radix;
			if (sign)
				goto handle_sign;
			goto handle_nosign;
		case 's':
			p = va_arg(ap, char *);
			if (p == NULL)
				p = "(null)";
			if (!dot)
				n = (int) strlen (p);
			else
				for (n = 0; n < dwidth && p[n]; n++)
					continue;

			width -= n;

			if (!ladjust && width > 0)
				while (width--)
					PCHAR(padc);
			while (n--)
				PCHAR(*p++);
			if (ladjust && width > 0)
				while (width--)
					PCHAR(padc);
			break;
		case 't':
			tflag = 1;
			goto reswitch;
		case 'u':
			base = 10;
			goto handle_nosign;
		case 'X':
			upper = 1;
		case 'x':
			base = 16;
			goto handle_nosign;
		case 'y':
			base = 16;
			sign = 1;
			goto handle_sign;
		case 'z':
			zflag = 1;
			goto reswitch;
handle_nosign:
			sign = 0;
			if (jflag)
				num = va_arg(ap, uintmax_t);
			else if (qflag)
				num = va_arg(ap, u_quad_t);
			else if (tflag)
				num = va_arg(ap, ptrdiff_t);
			else if (lflag)
				num = va_arg(ap, u_long);
			else if (zflag)
				num = va_arg(ap, size_t);
			else if (hflag)
				num = (u_short)va_arg(ap, int);
			else if (cflag)
				num = (u_char)va_arg(ap, int);
			else
				num = va_arg(ap, u_int);
			goto number;
handle_sign:
			if (jflag)
				num = va_arg(ap, intmax_t);
			else if (qflag)
				num = va_arg(ap, quad_t);
			else if (tflag)
				num = va_arg(ap, ptrdiff_t);
			else if (lflag)
				num = va_arg(ap, long);
			else if (zflag)
				num = va_arg(ap, ssize_t);
			else if (hflag)
				num = (short)va_arg(ap, int);
			else if (cflag)
				num = (char)va_arg(ap, int);
			else
				num = va_arg(ap, int);
number:
			if (sign && (intmax_t)num < 0) {
				neg = 1;
				num = -(intmax_t)num;
			}
			p = ksprintn(nbuf, num, base, &tmp, upper);
			if (sharpflag && num != 0) {
				if (base == 8)
					tmp++;
				else if (base == 16)
					tmp += 2;
			}
			if (neg)
				tmp++;

			if (!ladjust && padc != '0' && width
			    && (width -= tmp) > 0)
				while (width--)
					PCHAR(padc);
			if (neg)
				PCHAR('-');
			if (sharpflag && num != 0) {
				if (base == 8) {
					PCHAR('0');
				} else if (base == 16) {
					PCHAR('0');
					PCHAR('x');
				}
			}
			if (!ladjust && width && (width -= tmp) > 0)
				while (width--)
					PCHAR(padc);

			while (*p)
				PCHAR(*p--);

			if (ladjust && width && (width -= tmp) > 0)
				while (width--)
					PCHAR(padc);

			break;
		default:
			while (percent < fmt)
				PCHAR(*percent++);
			/*
			 * Since we ignore an formatting argument it is no
			 * longer safe to obey the remaining formatting
			 * arguments as the arguments will no longer match
			 * the format specs.
			 */
			stop = 1;
			break;
		}
	}
#undef PCHAR
}

static void write_buf
     (void *arg,const char *b,size_t len)
{
	char **x;
	x= arg;
    memcpy(*x,b,len);
    *x += len;
}

static void ignore
     (void *arg, const char *b, size_t len)
{
}

_MPRINTF_PROTO char *vmprintf(const char *fmt, va_list ap)
{
    va_list ap2;
    size_t size;
    char *ret,*x;
    int eno;

    eno= errno;

    va_copy(ap2,ap);
    size= vgprintf(ignore, NULL, fmt, ap2);
    va_end(ap2);

    x= ret= malloc(size+1);
    if (ret)
    {
        errno= eno;
        vgprintf(write_buf, (void*) &x,fmt, ap);
		ret[size] = 0;
    }
    
    return ret;
}


#define _BUFFER_PROTO static

typedef struct buffer buffer_t;
#ifndef _BUFFER_PROTO
#define _BUFFER_PROTO 
#endif

_BUFFER_PROTO buffer_t *buffer_new(size_t blksize);
_BUFFER_PROTO void buffer_free(buffer_t *B, void **Rd, size_t *Rl);
_BUFFER_PROTO void buffer_collect(buffer_t *B, void **Rd, size_t *Rl);
_BUFFER_PROTO void buffer_write(buffer_t *B, void *d, size_t l);
_BUFFER_PROTO void buffer_append(buffer_t *dst, buffer_t *src_f);
_BUFFER_PROTO size_t buffer_read(buffer_t *B, void *d, size_t l);

struct buffer_node { 
  struct buffer_node *next;
  size_t len,size,start;
  unsigned char data[1];
};

struct buffer {
  struct buffer_node *first, *last;
  size_t blksize;
};



_BUFFER_PROTO buffer_t *buffer_new(size_t blksize)
{
  buffer_t *B;
  B= malloc(sizeof(*B));
  B->first= B->last= NULL;
  B->blksize= blksize;
  return B;
}



_BUFFER_PROTO void buffer_collect(buffer_t *B, void **Rd, size_t *Rl)
{
  size_t S;
  struct buffer_node *P;
  unsigned char *R;

  if (!Rd && !Rl) return ;

  S= 0;
  for(P=B->first;P;P=P->next) S+= P->len;

  if (S==0) { if (Rd) *Rd= NULL; if (Rl) *Rl= 0; return ; }
  if (!Rd) { *Rl= S; return ; }

  R= malloc(S);
  S= 0;
  for(P=B->first;P;P=P->next)
  {
    memcpy(R+S, P->data+P->start, P->len);
    S+= P->len;
  }

  *Rd= (void*) R;
  if (Rl) *Rl= S;
}

_BUFFER_PROTO void buffer_free(buffer_t *B, void **Rd, size_t *Rl)
{
  struct buffer_node *P,*N;
  if (Rd || Rl) buffer_collect(B, Rd, Rl);
  for(P=B->first;P;P=N)
  {
    N= P->next;
    free(P);
  }
  free(B);
}

_BUFFER_PROTO void buffer_write(buffer_t *B, void *d, size_t l)
{
  struct buffer_node *P;
  size_t cp;

  while(l)
  {
    P= B->last;
    if (!P || (P->start+P->len)==P->size)
    {
      P= malloc(sizeof(*P)+B->blksize-1);
      P->len= 0;
      P->next= NULL;
      P->size= B->blksize;
      P->start= 0;
      if (!B->last) { B->last= B->first= P; }
      else { B->last->next= P; B->last= P; }
    }

    cp= P->size-P->len-P->start;
    if (cp>l) cp= l; 

    memcpy(P->data+P->len+P->start, d, cp);
    P->len+= cp;
    l-= cp;
    d= (unsigned char*) d + cp;
  }
}

_BUFFER_PROTO void buffer_append(buffer_t *dst, buffer_t *src_f)
{
  if (src_f->first!=NULL)
  {
    if (dst->first==NULL)
    {
      dst->first= src_f->first;
      dst->last= src_f->last;
    }
    else
    {
      dst->last->next= src_f->first;
      dst->last= src_f->last;
    }
  }
  free(src_f);
}



_BUFFER_PROTO size_t buffer_read(buffer_t *B, void *d, size_t l)
{
  size_t total;
  struct buffer_node *P;
  size_t cp;
  total= 0;

  while(l)
  {
    P= B->first;
    if (!P) break;  
    if (P->len==0)
    {
      B->first= P->next;
      if (!B->first) B->last= NULL;
      free(P);
      continue;
    } 
    cp= P->len;
    if (cp>l) cp= l;
    memcpy(d, P->data+P->start, cp);
    P->start+= cp;
    P->len-= cp;
    total+= cp;
    d= (unsigned char*) d + cp;
  }
  return total;
}

typedef struct { int left,right,top,bottom; } rect_t;
typedef struct { int x,y,w,h; } rrect_t; // relative rectangle

/***********************************************************************
 * A rectangle queue is used for storing the list of visible widgets   *
 * behind the current screen.                                          *
 ***********************************************************************/
typedef struct
{
  int head, tail,size;
  rect_t *data;
} rectque_t;

#define GFXceil(X,Y)  (((X)+((Y)-1))/(Y))

static int range_test(int v, int s, int w) { return v>=s && v<(s+w); }

static int imin2(int a,int b) { return a < b ? a : b; }
static int imax2(int a,int b) { return a > b ? a : b; }

static double dmin2(double a,double b) { return a<b?a:b; }
static double dmax2(double a,double b) { return a>b?a:b; } 

static int flag2(int a,int b)
{
  int S= 0;
  if (a) S|= 2;
  if (b) S|= 1;
  return S;
}

static int rectangle_intersect(rect_t *A, rect_t *B, rect_t *K)
{
  int N;
  *K= *A;
  N= 0;
  if (B->left>=A->left && B->left<=A->right) { K->left= B->left; N++; }
  if (B->right>=A->left && B->right<=A->right) { K->right= B->right; N++; }
  if (B->top>=A->top && B->top<=A->bottom) { K->top= B->top; N++; }
  if (B->bottom>=A->top && B->bottom<=A->bottom) { K->bottom= B->bottom; N++; }
  return N;
}

static int subtract_rectangle(rect_t *A,rect_t *B,rect_t *R)
{
  int N;
  N= 0;
  if (B->top<=A->bottom && B->top>=A->top)
  {
    *R= *A; R->bottom= B->top;
    R++; N++;
    A->top= B->top;
  }
  if (B->bottom<=A->bottom && B->bottom>=A->top)
  {
    *R= *A; R->top= B->bottom;
    R++; N++;
    A->bottom= B->bottom;
  }
  if (B->left<=A->right && B->left>=A->left)
  {
    *R= *A; R->right= B->left;
    R++; N++;
    A->left= B->left;
  }
  if (B->right<=A->right && B->right>=A->left)
  {
    *R= *A; R->left= B->right;
    R++; N++;
    A->right= B->right;
  }
  if (N) return N;
  if (B->right<A->left || B->left>A->right || B->top>A->bottom ||
      B->bottom<A->top) return 0;
  *R= *A;
  return 1;
}


static int in_rect(int rx,int ry,int rw,int rh, int x,int y)
{
  if (x<rx || y<ry || x>=(rx+rw) || y>=(ry+rh)) return 0;
  return 1;
}

static int rect_intersect
   (int *Rx, int *Ry, int *Rw, int *Rh,
      int tx, int ty, int tw, int th)
{
  int x0, x1, y0, y1;
  x0= *Rx; x1= x0+*Rw;
  y0= *Ry; y1= y0+*Rh;

  x0= imax2(x0, tx);   x1= imin2(x1, tx+tw);
  y0= imax2(y0, ty);   y1= imin2(y1, ty+th);
  *Rw= x1-x0; if (*Rw<=0) return 1;
  *Rh= y1-y0; if (*Rh<=0) return 1;
  *Rx= x0; *Ry= y0; return 0;
} 



static rectque_t *rectque_new()
{
  rectque_t *Q;
  Q= malloc(sizeof(*Q));
  Q->size= 20;
  Q->data= malloc(Q->size*sizeof(Q->data[0]));
  Q->head= Q->tail= 0;
  return Q;
}

static void* rectque_free(rectque_t *Q)
{
  free(Q->data);
  free(Q);
  return NULL;
}

static int rectque_len(rectque_t *Q)
{
  if (Q->head==Q->tail) return 0;
  if (Q->tail>Q->head) return Q->tail-Q->head;
  return Q->size-Q->head + Q->tail;
}

static void rectque_get(rectque_t *Q, rect_t *A)
{
  *A= Q->data[Q->head];
  Q->head= (Q->head+1)%Q->size;
}

static void rectque_put(rectque_t *Q, rect_t *A)
{
  if (rectque_len(Q)==Q->size-1)
  {
    rect_t *nd;
    int ns,i;
    ns= (Q->size+Q->size/2);
    nd= malloc(sizeof(nd[0])*ns);
    for(i=0;Q->head!=Q->tail;i++) rectque_get(Q, nd+i);
    nd[i++]= *A;
    Q->head= 0; Q->tail= i;
    free(Q->data);
    Q->data= nd;
    Q->size= ns;
  }
  else
  {
    Q->data[Q->tail]= *A;
    Q->tail= (Q->tail+1)%Q->size;
  }
}

static void rectque_putn(rectque_t *Q, rect_t *A, int N)
{
  int i;
  // I'm too lazy to do properly right now
  for(i=0;i<N;i++) rectque_put(Q,A+i);
}

static void rectque_get_data(rectque_t *Q, rect_t **D, int *N)
{
  rect_t *K;
  int i,j;
  if (Q->head==Q->tail) { *D= NULL; *N= 0; return  ; }
  K= malloc(sizeof(K[0])*rectque_len(Q));
  for(i=0,j=Q->head;j!=Q->tail;i=i+1,j=(j+1)%Q->size)
     K[i]= Q->data[j];
  *D= K;
  *N= i;
}


static void permute(int N, int *pP, void *aA, size_t size)
{
  int *T,*P,j,i,prev,next;
  unsigned char *A,t;

  T= malloc(sizeof(int)*N*2);
  P= T+N;
  A= (unsigned char*) aA;
  for(i=0;i<N;i++)
  {
    P[i]= pP[i];
    T[P[i]]= i;
  }

  for(i=0;i<N;i++)
  {
    if (T[i]==i) continue;
    next= T[i];
    prev= P[i];

    for(j=0;j<size;j++)
    { t= A[i*size+j];
      A[i*size+j]= A[prev*size+j];
      A[prev*size+j]= t; }

    T[prev]= next;
    P[next]= prev;
  }
  free(T);
}

static char *nstrdup(const char *k)
{
  if (k) return strdup(k);
    else return NULL;
}


static int utf8_encode_string
    (const uint32_t* wstr,size_t wlen, uint8_t *dst, size_t *dlen)
{
/*
  This is all from
  http://www.cl.cam.ac.uk/~mgk25/unicode.html

  TODO: what to do if the input has invalid
  unicode chars such as unassigned chars etc?
  pass them thru?
*/

/* Encoding Table
                  Start-Byte    
 Min Value    8421 8421   hex     Length
0x04000000    1111 110x   fc        6
0x00200000    1111 10xx   f8        5
0x00010000    1111 0xxx   f0        4
0x00000800    1110 xxxx   e0        3
0x00000080    110x xxxx   c0        2 */
  int i,k,len;
  uint32_t chr;
  unsigned char *odst;

  static const unsigned int lowerlimit[]=
    { 0x04000000, 0x00200000, 0x00010000, 0x00000800, 0x00000080 };
  static const unsigned char topbits []= { 0xfc, 0xf8, 0xf0, 0xe0, 0xc0 };
  const unsigned char lower6=  0x3f; // binary 0011 1111
  const unsigned char seventh= 0x80; // binary 1000 0000
  const unsigned int bit31=    0x80000000; 

  odst= dst;

  while(wlen)
  {
    chr= *(wstr++); wlen--;
    if (chr<=0x7f)   { *(dst++)= chr; continue; }
    if (chr & bit31) { *(dst++)= chr & ~bit31; continue; }

    for(i=0;i<4;i++) if (chr >= lowerlimit[i]) break;
    len= 4-i+2;

    for(k=len-1;k>0;k--)
    {
      dst[k]= seventh | (chr & lower6);
      chr >>= 6;
    }
    dst[k]= chr | topbits[i];
    dst+= len;
  }

  *dlen= dst-odst;
  return 0;
}

static void utf8_encode_single_char(uint32_t wide, char *dst)
{
  /* This assumes that dst is big enough, at least 6 bytes. */
  size_t L;
  L= 6;
  utf8_encode_string(&wide,1, (uint8_t*) dst, &L);
}

/*
UTF-8 Table
8421 8421
110x xxxx   2 bytes    mask= e0  result= c0   min= 80       vmask= 1f
1110 xxxx   3 bytes    mask= f0  result= e0   min= 800      vmask= 0f
1111 0xxx   4 bytes    mask= f8  result= f0   min= 10000    vmask= 07
1111 10xx   5 bytes    mask= fc  result= f8   min= 200000   vmask= 03
1111 110x   6 bytes    mask= fe  result= fc   min= 4000000  vmask= 01

10xx xxxx   cont. byte mask= c0  result= 80   vmask= 3f
*/
static int utf8_decode_string
  (const uint8_t *input, size_t ilen, uint32_t *output, size_t *olen)
{
   const uint8_t *end,*start;
   static const uint32_t bit31= (1<<31);
   size_t L;
   int have_errors;
   int i, contb;
   uint8_t C;
   uint32_t outc; uint32_t min;

   end= input+ilen;
   L= 0;
   have_errors= 0;

   if (ilen==0) goto done;

again:
   C= input[0];
   if ( (C&0x80) ==0)    { outc= C; input++; goto out; }
   if ( (C&0xe0) ==0xc0) goto seq2; 
   if ( (C&0xf0) ==0xe0) goto seq3;
   if ( (C&0xf8) ==0xf0) goto seq4;

   // Here, we don't care about the seq5 and seq6 stuff.
   // they are invalid anyway, either overlong or greater
   // then the unicode limit. We just treat them as unrecognized
   // bytes and output them.
invalid_byte:
   outc= input[0] | bit31;
   input++;
   have_errors= 1;
   goto out;

seq2: outc= C & 0x1f;  min=    0x80; contb= 1; goto do_cont;
seq3: outc= C & 0x0f;  min=   0x800; contb= 2; goto do_cont;
seq4: outc= C & 0x07;  min= 0x10000; contb= 3; goto do_cont;

do_cont:
   start= input;
   input++;
   for(i=0;i<contb;i++)
   {
     if ((input==end) || ( (input[0]&0xc0) !=0x80)) goto reject;
     outc= (outc << 6) | (input[0]&0x3f);
     input++;
   }

   if (outc<min || outc>0x10ffff) goto reject; 
   if (outc>=0xd800 && outc<=0xdfff) goto reject;
   if (outc==0xffff || outc==0xfffe) goto reject;

out:
   output[L++]= outc;
   if (input<end) goto again;
done:
   *olen= L;
   return have_errors;
reject:
   input= start;
   goto invalid_byte; 
}


static uint32_t utf8_decode_single_char(char *src)
{
  uint32_t *dst;
  size_t L;
  L= strlen(src);
  dst= alloca(sizeof(*dst)*(L+1));
  utf8_decode_string((uint8_t*) src, L, dst, &L);
  if (L==0) return ' ';
  return dst[0];
}



static void utf8_decode_alloc(char *str, uint32_t **Rwstr, uint32_t *Rwlen)
{
  uint32_t *S;
  size_t L,SL;
  L= strlen(str);
  S= malloc((L+1)*sizeof(uint32_t));
  utf8_decode_string((uint8_t*)str, L, S, &SL);
  *Rwstr= S;
  *Rwlen= (uint32_t) SL;
}


static int clip_for_rect(int *start, int *amount, int max)
{
  int end, am;
  am= *amount;
  if (am<=0) return 1;
  end= *start + am;
  if (end>max) end= max;
  if (*start<0) *start= 0;
  *amount= am= end - *start;
  return am<=0;
}

static void freeNnull(void *A)
{
  if (A) free(A);
}

static void *array_make1space
      (int pos,void **Rarray,int n, int *Rs, int eltsize)
{
  char *src, *dst;
  int cp;
  int s;

  src= (void*) *Rarray;
  s= *Rs;
  if (n<s)
  {
    dst= src;
  }
  else
  {
    if (s==0) s= 50;
         else s+= s/2;
    dst= malloc(s*eltsize);
  }

  if (dst!=src && src && pos)
    memcpy(dst, src, pos*eltsize);
  cp= n - pos;
  if (src && cp)
    memmove(dst+(pos+1)*eltsize, src+pos*eltsize, cp*eltsize);

  if (src && src!=dst) free(src);
  *Rarray= dst;
  *Rs= s;
  return dst+pos*eltsize;
}

static int clip_rrect(rrect_t *D, rrect_t *S)
{
  int XsE, YsE, XdE, YdE;
  XsE= S->x+S->w;   XdE= D->x+D->w;
  YsE= S->y+S->h;   YdE= D->y+D->h;

  if (XsE>XdE) S->w -= XsE-XdE; 
  if (YsE>YdE) S->h -= YsE-YdE; 

  if (S->x < D->x) { S->w -= D->x - S->x; S->x= D->x; }
  if (S->y < D->y) { S->h -= D->y - S->y; S->y= D->y; }

  if (S->w<=0 || S->h<=0) return 1;
  return 0;
}


/***********************************************************************
 *                            Path Handling                            *
 ***********************************************************************/

#define __PATH_PROTO static

static void strmove(char *dst,char *src)
{
  memmove(dst, src, strlen(src)+1);
}

static int remove_mid(char *path, char *pattern, int keep)
{
  char *k;
  k= strstr(path, pattern);
  if (!k) return 0;
  strmove(k+keep, k+strlen(pattern));
  return 1;
}

static int remove_prefix(char *path, char *pattern, int keep)
{
  int i;
  for(i=0;pattern[i];i++)
   if (path[i]!=pattern[i]) return 0;
  strmove(path+keep,path+i);
  return 1;
  
}

static char *has_suffix(char *path, char *sfx)
{
  int sL, pL,i;
  sL= (int) strlen(sfx);
  pL= (int) strlen(path);
  if (pL<sL) return NULL;
  path+= pL-sL;
  for(i=0;path[i];i++)
    if (path[i]!=sfx[i]) return NULL;
  return path;
}

static int remove_suffix(char *path,char *sfx,int keep)
{
  if (!(path=has_suffix(path,sfx))) return 0;
  path[keep]= 0;
  return 1;
}


__PATH_PROTO char *path_file(char *path)
{
  char *G;
  if (!path || !path[0]) return strdup("");
  path= strdup(path);
  while(0
    || remove_mid(path,"/./",1) 
    || remove_mid(path,"//",1)  
    || remove_suffix(path,"/.",1)
    || remove_suffix(path,"/..",1)
       );
  G= strrchr(path,'/');
  if (G)
    strmove(path, G+1);
  return path;
}

static int dotdotlast(char *path)
{
  char *k, *u,*z;
  k= has_suffix(path,"/..");
  if (!k) return 0;

  z= NULL;
  for(u=path;u<k;u++)
    if (u[0]=='/') z= u;
  if (!z) return 0;
  z[1]= 0;
  return 1;
}

static int dotdot(char *path)
{
  char *k, *u,*z;
  k= strstr(path,"/../");
  if (k==NULL) return 0;
  if (k==path) return remove_prefix(path,"/../",1);
  z= NULL;
  for(u=path;u<k;u++)
   if (u[0]=='/') z= u;
  if (!z) strmove(path, k+4); 
     else strmove(z+1, k+4);
  return 1;
}

__PATH_PROTO char *path_dir(char *path)
{
  char *G;
  if (!path || !path[0]) return strdup("");
  path= strdup(path);
  while(0
    || remove_mid(path,"/./",1) 
    || remove_mid(path,"//",1) 
    || remove_prefix(path,"/../",1)
    || remove_suffix(path,"/.", 1)
    || dotdot(path)
    || dotdotlast(path)
       );
  
  G= strrchr(path,'/');
  if (!G)           path[0]= 0; 
  else if (G==path) path[1]= 0; 
  else              G[0]= 0; 
  return path;
}

__PATH_PROTO char *path_wd()
{
#ifdef GFX_UNIX
  char *buf,*ret;
  size_t siz;
  siz= 1000;
again:
  buf= malloc(siz);
  if (getcwd(buf,siz)==NULL) 
  {
     if (errno==ERANGE) { siz+= siz/2; free(buf); goto again; }
     strcpy(buf,"/");
  }
  ret= strdup(buf);
  free(buf);
  return ret;
#else
	return strdup("/");
#endif
}

static char *pacom_fk(char *a,char* b)
{
  char *R;
  R= malloc(strlen(a)+strlen(b)+2);
  sprintf(R,"%s/%s",a,b);
  free(a);
  return R;
}

__PATH_PROTO char* path_absolute(char *path)
{
  if (!path || !path[0]) return strdup("");

  if (path[0]=='/') path= strdup(path);
              else  path= pacom_fk(path_wd(), path);

  while(0
    || remove_mid(path,"/./",1) 
    || remove_mid(path,"//",1) 
    || remove_prefix(path,"/../",1)
    || remove_suffix(path,"/.",1)
    || dotdot(path)
    || dotdotlast(path)
       );

  return path;
}

__PATH_PROTO char *path_relative(char *path, char *pivot)
{
  int i,j,k;
  int count;
  char *result;

  path= path_absolute(path);
  pivot= path_absolute(pivot);

  for(i=0;path[i];i++)
    if (path[i]!=pivot[i]) break;

  if (!path[i]) 
  {
    result= path_file(path);
    goto done;
  }

  count= 1;
  for(j=i-1;j>0;j--) if (path[j]=='/') break;

  /* from j to end, count how many directories you have.. */

  count= 0;
  for(k=j+1;path[k];k++) if (path[k]=='/') count++;

  result= malloc(strlen(path+j) + count*3 + 1);
  result[0]= 0;
  for(i=0;i<count;i++) strcat(result,"../");
  strcat(result,path+j+1);

done:
  free(path);
  free(pivot);
  return result;
}

__PATH_PROTO char *path_child(char *path,char *child)
{
  char *R;
  R= malloc(strlen(path) + strlen(child) + 2);
  sprintf(R,"%s/%s",path,child);
  return R;
}


/***********************************************************************
 *                             Object Map                              *
 ***********************************************************************
  This part implements an object map based on critbit trees.
 ***********************************************************************/

#define WOMAP_PROTO static

typedef struct { void *root; } womap_t;

WOMAP_PROTO void* womap_find(womap_t *map,uint32_t key);
WOMAP_PROTO void* womap_insert(womap_t* map, uint32_t key, void *data);
WOMAP_PROTO void *womap_remove(womap_t* map, uint32_t key);
WOMAP_PROTO void *womap_destroy(womap_t *map);
WOMAP_PROTO womap_t  *womap_new();

typedef struct
{
  union { struct { void* child[2];           } inte;  
          struct { void *data; uint32_t key; } leaf; } d;
  unsigned internal:1, bit:5;
} womap_node_t;

 // careful, qumap also uses the same things
#ifndef Q_KEY
#define Q_KEY(X)    ((X)->d.leaf.key)
#define Q_DATA(X)   ((X)->d.leaf.data)
#define Q_CHLD(X,Y) ((X)->d.inte.child[Y])
#endif


static inline int womap_dir(uint32_t key,womap_node_t *node);
static inline womap_node_t* womap_leaf_node(uint32_t key, void *data);
static inline womap_node_t* womap_internal_node(uint32_t bit);
static uint8_t womap_crit_bit(uint32_t X, uint32_t Y);

static inline int womap_dir(uint32_t key,womap_node_t *node)
{
  return (key >> node->bit)&1;
}

static inline womap_node_t* womap_leaf_node(uint32_t key, void *data)
{
  womap_node_t *N;
  N= malloc(sizeof(*N));
  N->internal= 0;
  Q_KEY(N)= key;
  Q_DATA(N)= data; 
  return N;
}

static inline womap_node_t* womap_internal_node(uint32_t bit)
{
  womap_node_t *N;
  N= malloc(sizeof(*N));
  N->internal= 1;
  N->bit= bit;
  return N;
}

static uint8_t womap_crit_bit(uint32_t X, uint32_t Y)
{
  uint32_t i,j;
  static const uint32_t B[]=
  {
    0x55555555,
    0x33333333,
    0x0f0f0f0f,
    0x00ff00ff,
    0x0000ffff,
  };
  uint32_t A;
  A= X ^ Y;
  
  for(i=1;i<32;i<<=1) A |= (A<<i);
  A = ~A;
  for(j=0,i=1;i<32;i<<=1,j++) A= (A&B[j]) + ((A>>i)&B[j]);
  return A;
}


WOMAP_PROTO void* womap_find(womap_t *map,uint32_t key)
{
  womap_node_t *node;

  node= map->root;
  if (!node) return NULL;

  while(node->internal) node= Q_CHLD(node, womap_dir(key,node));

  if (Q_KEY(node)==key) return Q_DATA(node);
                          else return NULL;
}


WOMAP_PROTO void* womap_insert(womap_t* map, uint32_t key, void *data)
{
  womap_node_t *node_elt;

  if (!map->root)
  {
     map->root= womap_leaf_node(key,data);
     return 0;
  }

  node_elt= map->root;
  while(node_elt->internal)
    node_elt= Q_CHLD(node_elt, womap_dir(key,node_elt));

  
  uint32_t elt;
  int dir_elt;
  uint8_t bit;
  womap_node_t *node_key;

  elt= Q_KEY(node_elt);

  if (elt==key) return Q_DATA(node_elt);

  bit= womap_crit_bit(elt,key);

  dir_elt= (elt>>bit)&1;

  node_key= womap_internal_node(bit);
  Q_CHLD(node_key,1-dir_elt)= womap_leaf_node(key,data);

  womap_node_t *parent;
  womap_node_t *node;

  parent= 0;
  node= map->root;

  while(node->internal)
  {
     if (node->bit>bit) break;
     parent= node;
     node= Q_CHLD(node, womap_dir(key,node));
  }

  Q_CHLD(node_key,dir_elt)= node;
  if (parent) { if (Q_CHLD(parent,0)==node)  Q_CHLD(parent,0)= node_key;
                                       else  Q_CHLD(parent,1)= node_key; }
  else map->root= node_key;
  return 0;
}

WOMAP_PROTO void *womap_remove(womap_t* map, uint32_t key)
{
  void *retv;

  if (!map->root) return NULL;

   
  womap_node_t *node;
  womap_node_t *sibling;
  womap_node_t *parent;
  womap_node_t *gparent;

  node= map->root;
  gparent= 0;
  sibling= 0;
  parent= 0;

  while(node->internal)
  {
    int dir;
    dir= womap_dir(key, node);
    gparent= parent;
    parent= node;
    sibling= Q_CHLD(node, 1-dir);
    node= Q_CHLD(node, dir);
  }

  if (Q_KEY(node)!=key) return NULL;
  retv= Q_DATA(node);
  free(node);

  if (!gparent)
  {
     map->root= sibling;
     if (parent) free(parent);
     return retv;
  }

  if (Q_CHLD(gparent,0)==parent) Q_CHLD(gparent,0)= sibling;
                            else Q_CHLD(gparent,1)= sibling;
  free(parent);

  return retv;
}

WOMAP_PROTO void *womap_destroy(womap_t *map)
{
  womap_node_t *node;
  womap_node_t *parent;
  womap_node_t *gparent;
  womap_node_t *sibling;
  void *retv;

  if (!map->root) return 0;

  parent= 0;
  gparent= 0;
  sibling= 0;
  node= map->root;
  while(node->internal)
  {
    gparent= parent;
    parent= node;
    sibling= Q_CHLD(node, 1);
    node= Q_CHLD(node,0);
  }
  retv= Q_DATA(node);
  free(node);
  if (!gparent) 
  {
    map->root= sibling;
    if (parent) free(parent);
    return retv;
  }
  if (Q_CHLD(gparent,0)==parent)
    Q_CHLD(gparent,0)= sibling;
  else
    Q_CHLD(gparent,1)= sibling;

  free(parent);

  return retv;
}

WOMAP_PROTO womap_t  *womap_new()
{
  womap_t *s;
  s= malloc(sizeof(*s));
  s->root= 0;
  return s;
}

