
/***********************************************************************
 * #WIN                                                                *
 *                            WINDOWS SUBSYSTEM                        *
 *                                                                     *
 ***********************************************************************
   For any new port, 3 functions need to be written:
     - gfx_init
     - gfx_update
     - get_event
     - getdpi
   That's it. 
 ***********************************************************************/

#ifdef GFX_WIN
static unsigned int *x_image;
static HDC x_DC;
static HWND x_window;
static HBITMAP x_bitmap;
static gfx_event_t* EVRET;
static int gui_have_event;
static HACCEL x_accel_table;

#define x_red_mask  0xff0000
#define x_green_mask 0x00ff00
#define x_blue_mask 0x0000ff
#define x_red_shift  16
#define x_green_shift 8
#define x_blue_shift 0

static int getdpi() { return 96; } // FIXME

static LRESULT CALLBACK WndProc(HWND hwnd, UINT msg,
                                WPARAM wParam, LPARAM lParam) 
{
  HDC hdc;
  PAINTSTRUCT ps;
  int x,y,w,h;
  int k,t;
  int delta;
  static char keybuf[10];

  switch(msg) {
  case WM_CREATE:
    break;
  case WM_PAINT:
    hdc = BeginPaint(hwnd, &ps);
    x= ps.rcPaint.left;
    y= ps.rcPaint.top;
    w= ps.rcPaint.right-x+1; h= ps.rcPaint.bottom-y+1;
    BitBlt(hdc,x,y,w,h,x_DC,x,y,SRCCOPY);
    EndPaint(hwnd, &ps);
    break;
  case WM_DESTROY:
    PostQuitMessage(0);
    return 0;
  case WM_CLOSE:
    EVRET->type= GFXEV_QUIT; 
    gui_have_event= 1;
    return 0;
  case WM_LBUTTONDBLCLK: k = 1; t= GFXEV_DOUBLE_CLICK; goto mouse;
  case WM_LBUTTONDOWN: k = 1;  t = GFXEV_BUTTON_PRESS; goto mouse;
  case WM_LBUTTONUP: k = 1; t = GFXEV_BUTTON_RELEASE; goto mouse;
  case WM_RBUTTONDBLCLK: k = 2; t = GFXEV_DOUBLE_CLICK; goto mouse;
  case WM_RBUTTONDOWN: k = 2; t = GFXEV_BUTTON_PRESS; goto mouse;
  case WM_RBUTTONUP: k = 2; t = GFXEV_BUTTON_RELEASE; goto mouse;
  case WM_MBUTTONDBLCLK: k = 3; t = GFXEV_DOUBLE_CLICK; goto mouse;
  case WM_MBUTTONDOWN: k = 3; t = GFXEV_BUTTON_PRESS; goto mouse;
  case WM_MBUTTONUP: k = 3; t = GFXEV_BUTTON_RELEASE; goto mouse;
  case WM_MOUSEWHEEL: 
    delta = GET_WHEEL_DELTA_WPARAM(wParam); 
    if (delta < 0) k = 5; else k = 4;
    t = GFXEV_BUTTON_PRESS; 
    goto mouse;
  case WM_MOUSEMOVE: k = 0; t = GFXEV_POINTER_MOVE; goto mouse;
  case WM_KEYDOWN: t= GFXEV_KEY_PRESS; goto key;
  case WM_KEYUP: t = GFXEV_KEY_RELEASE; goto key;
  case WM_CHAR: k = GFXEV_KEY_PRESS; goto chr;
  }
default_handling:
  return DefWindowProcW(hwnd, msg, wParam, lParam);
mouse:
  // FIXME: also report modifier keys such as ctrl etc.
  // this is not done on other platforms, so I'll avoid it
  // for now. wParam contains the necessary information
  // if I want to implement it later. look for 
  //  
  // GET_KEYSTATE_WPARAM
  //
  EVRET->type = t; EVRET->button = k;
  EVRET->x = GET_X_LPARAM(lParam);
  EVRET->y = GET_Y_LPARAM(lParam);
  gui_have_event = 1;
  return 0;
key:
  EVRET->type = t; 
  switch (wParam)
  {
  case VK_LEFT: k = GFX_K_LEFT; break;
  case VK_RIGHT: k = GFX_K_RIGHT; break;
  case VK_HOME: k = GFX_K_HOME; break;
  case VK_END: k = GFX_K_END; break;
  case VK_UP: k = GFX_K_UP; break;
  case VK_DOWN: k = GFX_K_DOWN; break;
  case VK_PRIOR: k = GFX_K_PAGEUP; break;
  case VK_NEXT: k = GFX_K_PAGEDOWN; break;
  case VK_INSERT: k = GFX_K_INSERT; break;
  case VK_DELETE: k = GFX_K_DELETE; break;
  case VK_ESCAPE: k = GFX_K_ESCAPE; break;
  case VK_TAB: k = GFX_K_TAB; break;
  case VK_RETURN: k = GFX_K_RETURN; break;
  case VK_BACK:k = GFX_K_BACKSPACE; break;
  default:
    goto default_handling;
  }
  EVRET->key = k;
  gui_have_event = 1;
  return 0;
chr:
  switch (wParam)
  {
  case '\b': wParam = VK_BACK; t = GFXEV_KEY_PRESS; goto key;
  case '\n':
  case '\r': wParam = VK_RETURN; t = GFXEV_KEY_PRESS; goto key;
  case '\t': wParam = VK_TAB; t = GFXEV_KEY_PRESS; goto key;
  }
  EVRET->type = GFXEV_KEY_PRESS;
  EVRET->key = GFX_K_PRT;
  EVRET->wide = wParam;
  memset(keybuf, 0, sizeof(keybuf));
  utf8_encode_single_char(wParam, keybuf);
  EVRET->string = keybuf;
  gui_have_event = 1;
  return 0;
}

static int get_event(gfx_event_t* Revent)
{
  int R;
  MSG msg;
again:
  R = GetMessage(&msg, x_window, 0, 0);
  if (R <= 0) return 0;
  gui_have_event = 0;
  EVRET = Revent;
  if (!TranslateAccelerator(x_window, x_accel_table, &msg))
  {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }
  if (!gui_have_event) goto again;
  gui_have_event = 0;
  return 1;
}

static void gfx_update(int x, int y, int width, int height)
{  
  clip_for_rect(&x, &width, x_window_width);
  clip_for_rect(&y, &height, x_window_height);
  if (width <= 0 || height <= 0) return;
  RECT u;
  u.top = y;
  u.left = x;
  u.right = x + width - 1;
  u.bottom = y + height - 1;
  InvalidateRect(x_window, &u, FALSE);
}


static void gfx_start(HINSTANCE instance, int w, int h, char *window_title)
{
  BITMAPINFO bi;
  BITMAPINFOHEADER* BH;
  int i;
  WNDCLASS wc;
  RECT cla;

  static ACCEL x_accels[] =
  {
    { FVIRTKEY | FCONTROL, 'W' , WM_CLOSE }
  };

  memset(&wc, 0, sizeof(wc));
  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpszClassName = "gfxStar";
  wc.hInstance     = instance;
  wc.hbrBackground = GetSysColorBrush(COLOR_3DFACE);
  wc.lpfnWndProc   = WndProc;
  wc.hCursor       = LoadCursor(0, IDC_ARROW);

  RegisterClass(&wc);
  if (w==0 || h==0)
  {
    RECT wa;
    if (SystemParametersInfo(SPI_GETWORKAREA, 0, &wa, 0))
    {
      if (!w) w= wa.right - wa.left;
      if (!h) h= wa.bottom - wa.top;
    }
    else
    {
      if (!w) w= 400; 
      if (!h) h= 400;
    }
  }
printf("window title: %s\n", window_title);

  x_window= CreateWindow(wc.lpszClassName, window_title,
          WS_CAPTION | WS_MINIMIZEBOX | WS_SYSMENU | WS_VISIBLE,
         0, 0, w, h, NULL, NULL, instance, NULL);
  if (GetClientRect(x_window,&cla))
  {
    x_window_width= cla.right;
    x_window_height= cla.bottom;
  }
  else
  {
    x_window_width= w;
    x_window_height= h;
  }

  x_DC = CreateCompatibleDC(GetDC(x_window));
  memset(&bi, 0, sizeof(bi));
  BH = &bi.bmiHeader;
  BH->biSize = sizeof(*BH);
  BH->biWidth = x_window_width;
  BH->biHeight = -x_window_height;
  BH->biPlanes = 1;
  BH->biBitCount = 32;
  BH->biCompression = BI_RGB;
  x_bitmap = CreateDIBSection(x_DC, &bi, DIB_RGB_COLORS, 
                              (void**)&x_image, NULL, 0);
  x_image_data = malloc(sizeof(x_image_data[0]) * x_window_height);
  for (i = 0; i < x_window_height; i++) 
      x_image_data[i] = x_image + i * x_window_width;
  SelectObject(x_DC, x_bitmap);

  x_accel_table = CreateAcceleratorTable(x_accels, 
                      sizeof(x_accels) / sizeof(x_accels[0]));
  init_data();
}

void gfx_localdata(const unsigned char**,size_t*);

void gfx_init(HINSTANCE instance)
{
  const unsigned char *module, *module_end; size_t module_len;
  int window_width, window_height;
  char *window_title;
  gfx_localdata(&module, &module_len);
  module_end= module + module_len;
  window_title= gui_parse_string(&module);
  window_width= gui_parse_int(&module);
  window_height= gui_parse_int(&module);
  gfx_data(module);
  gfx_start(instance,window_width, window_height, window_title);
}
#endif
