#ifdef GFX_ANDROID
#ifndef GFX_ANDROID_RELEASE

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>


typedef struct { int fd; char host[200]; int port; } sock_t;

static int sock_server(sock_t **RS,int port);
static int sock_accept(sock_t **RS,sock_t *S);
static int sock_connect(sock_t **RS, uint32_t host, int port);
static int sock_close(sock_t *S);

#ifndef SOCK_CLOEXEC
#ifdef O_CLOEXEC
#define SOCK_CLOEXEC O_CLOEXEC
#else
#define SOCK_CLOEXEC 0
#endif
#endif 

static void sock_error(sock_t *S, const char *fmt, ...);



static int sock_connect(sock_t **RS, uint32_t host, int port)
{
  struct sockaddr_in addr;
  sock_t *S;

  S= calloc(1,sizeof(*S));
  *RS= S;

  S->port= port;
#if 0
  strncpy(S->host,host, sizeof(S->host));
  S->host[sizeof(S->host)-1]= 0;
#endif

  S->fd= socket(AF_INET, SOCK_STREAM, 0);
  if (S->fd<0) 
  {
    sock_error(S,"can't create socket: %s\n", strerror(errno));
    return 1;
  }

  memset(&addr, 0, sizeof(addr));
  addr.sin_family= AF_INET;
  addr.sin_addr.s_addr= host;
  addr.sin_port= htons(port);
  if (connect(S->fd, (struct sockaddr*) &addr, sizeof(addr))<0)
  {
    sock_error(S,"can't connect to %s: %s\n", host, strerror(errno));
    goto Fclose;
  }
  return 0;

Fclose:
  close(S->fd);
  S->fd= -1;
  return 1;
}

static int sock_close(sock_t *S)
{
  if (S->fd>=0) { shutdown(S->fd, SHUT_RDWR); close(S->fd); }
  memset(S,0, sizeof(*S));
  S->fd= -1;
  free(S);
  return 0;
}

static void sock_erro(sock_t *S, const char *data, int len)
{
  fwrite(data, 1, len, stderr);
}

static void sock_error(sock_t *S, const char *fmt, ...)
{
  int i;
  char buf[200];
  char *v;
  va_list ap;
  va_start(ap,fmt);

again:
  for(i=0;fmt[i];i++) if (fmt[i]=='%') break;

  if (i) sock_erro(S, fmt, i);
  if (!fmt[i]) goto done ;
  i++;
  switch(fmt[i])
  {
    case 0: sock_erro(S, "%", 1); goto done;
    case '%': sock_erro(S, "%", 1); break;
    case 's': v= va_arg(ap,char*); sock_erro(S, v, strlen(v)); break;
    case 'd': i= va_arg(ap,int); sprintf(buf,"%d", i);
              sock_erro(S, buf, strlen(buf)); break;
    default: sock_erro(S, fmt+i, 1); break;
  }
  fmt+= i+1;
  goto again;
done:
  va_end(ap);
}

static sock_t *losock;

static void locon(char* dst)
{
  int K;
  int b= 0;
  K= 0;
  while(dst[0] && b<32)
  {
    K |= atoi(dst) << b; b+= 8;
    while(dst[0]>='0' && dst[0]<='9') dst++;
    if (dst[0]=='.') dst++;
  }
  if (sock_connect(&losock, K, 4218)) losock= NULL;
}

static void logg(const char *fmt, ...)
{
  va_list ap;
  if (!losock) 
  {
    va_start(ap, fmt);
    __android_log_vprint(ANDROID_LOG_VERBOSE, "gfxApp", fmt, ap);
    va_end(ap);
  }
  else
  {
    char buf[2000];
    va_start(ap, fmt);
    vsnprintf(buf,sizeof(buf),fmt,ap);
    va_end(ap);
    send(losock->fd, buf, strlen(buf),0);
  }
}

#else

static void locon(char *addr) { }
static void logg(const char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  __android_log_vprint(ANDROID_LOG_VERBOSE, "gfxApp", fmt, ap);
  va_end(ap);
}

#endif
#endif

