
static uint32_t  fwmenu_bg_color= 0xffffff,
                 fwmenu_sep_color= 0x888888,
                 fwmenu_fg_color= 0x000000,
                 fwmenu_scrpad= 20,
                 fwmenu_intpad= 5;


int gfx_menu(int font,int nentries, TUPiis_t *entries)
{
  int i;
  int R;
  int maxtextw ,maxtexth;

  colorpix_t dbuf;
  graypix_t subind;
  graypix_t text;

  int maxwindoww, maxwindowh,my,dy,window_x,window_y,window_w,window_h;

  if (nentries<=0) return -1;

  static const unsigned char GMfwmenu_sub[]= {
    0x61,0x3c,0x00,0x00,0x00,0x00,0x4b,0x80,0x48,0x00,
    0x00,0x00,0x00,0x41,0x80,0x51,0x02,0x00,0x00,0x00,
    0x36,0x7f,0x59,0x04,0x00,0x00,0x00,0x2c,0x7d,0x56,
    0x00,0x00,0x00,0x2c,0x7d,0x5c,0x00,0x00,0x36,0x7f,
    0x60,0x07,0x00,0x41,0x80,0x56,0x03,0x00,0x48,0x80,
    0x4b,0x01,0x00,0x00,0x60,0x3e,0x00,0x00,0x00,0x00,
  };
  static const unsigned int GMWfwmenu_sub= 6, GMHfwmenu_sub=10;

  maxtextw= 0;
  maxtexth= 0;
  for(i=0;i<nentries;i++)
  {
    int w,h;
    gfx_text_size(font,entries[i].sv,&w,&h);
    maxtextw= imax2(w, maxtextw);
    maxtexth= imax2(h, maxtexth);
  }

  maxwindoww= x_window_width - 2*fwmenu_scrpad;
  maxwindowh= x_window_height - 2*fwmenu_scrpad;

  unsigned int *img, img_w, img_h;
  int ipad= maxtexth / 10;
  int item_h= maxtexth + ipad;

  img_w= imin2(maxwindoww, maxtextw+GMWfwmenu_sub+1+2*fwmenu_intpad);
  img_h= nentries*item_h;
  img= malloc(img_w*img_h*sizeof(uint32_t));

  for(i=0;i<img_w*img_h;i++) img[i]= fwmenu_bg_color;

  dbuf.w= img_w;
  dbuf.h= img_h;
  dbuf.stride= dbuf.w;
  dbuf.data= img;

  subind.w= GMWfwmenu_sub;
  subind.h= GMHfwmenu_sub;
  subind.stride= subind.w;
  subind.data= (unsigned char*) GMfwmenu_sub;

  int y, avail_width;
  y= 0;
  avail_width= img_w - 2*fwmenu_intpad - GMWfwmenu_sub-1;
  for(i=0;i<nentries;i++)
  {
    text_to_graypix(font, entries[i].sv, &text, NULL);
    text.w= imin2(avail_width, text.w); // this is safe, we still have
                                        // stride and we don't need the
                                        // text object later on.
    blend_graypix(fwmenu_fg_color, text, dbuf, 
                   fwmenu_intpad,
                   y+(item_h-text.h)/2);
    free(text.data);

    if (entries[i].iv2)
       blend_graypix(fwmenu_fg_color, subind, dbuf,
                     dbuf.w-fwmenu_intpad-GMWfwmenu_sub,
                     y + (text.h-GMHfwmenu_sub)/2);
    y+= item_h;
    colorpix_hline(&dbuf, fwmenu_intpad, img_w-fwmenu_intpad,y-1,
                                               fwmenu_sep_color);
  }

  window_w= img_w;
  window_h= imin2(maxwindowh, img_h);

  window_x= (x_window_width - window_w)/2;
  window_y= (x_window_height - window_h)/2;

  int pos, dirty, dragging, down;
  gfx_event_t event;
  uint32_t *backstore;

  R= -1;
  down= dragging= 0; dirty= 1;
  pos= 0;

  backstore= get_server_pix(window_x-2,window_y-2,window_w+4,window_h+4);

  gfx_draw_rect(0,window_x-1,window_y-1,window_w+2,window_h+2, 0x888888);
  gfx_draw_rect(GFX_UPDATE,window_x-2,window_y-2,window_w+4,window_h+4, 0x00);

  while(1)
  {
    if (dirty)
    {
     gfx_put(GFX_UPDATE, img, img_w, img_h,
                         0, pos, window_x, window_y,
                        img_w, window_h);
                         
     dirty= 0;
    }
    if (!get_event(&event)) continue;
    switch(event.type)
    {
    case GFXEV_BUTTON_PRESS:
        if (event.button!=1) continue;
        if (event.x<window_x || event.x>=(window_x+window_w) ||
            event.y<window_y || event.y>=(window_y+window_h)) goto cancelled;
        my= event.y;
        down= 1;
        dragging= 0;
        break;
    case GFXEV_BUTTON_RELEASE:
        if (event.button!=1) continue;
        down= 0;
        if (dragging)
        {
          dragging= 0;
          break;
        }
        R= (my - window_y + pos)/item_h;
        goto done;
    case GFXEV_POINTER_MOVE:
        if (!down || window_h>=img_h) break;
        if (!dragging)
        {
          if (abs(event.y-my)<=20) break;
          dragging= 1;
        }
        dy= event.y - my;
        if (!dy) break;
        my= event.y;
        if (dy<0) {  // dragging finger upwards
          if (pos+window_h>=img_h) break;
          pos -= dy; 
          if (pos+window_h>img_h) pos= img_h-window_h;
        } else
        {
          if (pos==0) break;
          pos-= dy;
          if (pos<0) pos= 0;
        }
        dirty= 1;
        break;
    }
  }

cancelled:

  R= -1;
  while(1)
  {
    if (!get_event(&event)) continue;
    switch(event.type)
    {
    case GFXEV_BUTTON_RELEASE:
        if (event.button==1) goto done;
    }
  }

done:
  free(img);
  put_server_pix(backstore, window_x-2,window_y-2,window_w+4,window_h+4);
  free(backstore);
  gfx_update(window_x-2,window_y-2,window_w+4,window_h+4);
  if (R>=0 && R<nentries)
     return entries[R].iv1; 
  else return -1;
}
