/***********************************************************************
 *                                                                     *
 *                             User Icons                              *
 *                                                                     *
 ***********************************************************************
  User icons are simply stored in an array. Indices to this array are
  mentioned in either widget creation functions or list insert functions.

  Each icon is a set of images, one for class of screen resolutions. 
  The range of valid resolutions for each image is determined by the
  user, given as a numerical value. 

  Once an icon is loaded, it's converted to user image format and
  then used from that raw data.

  In the module description, icons are represented as follows:
    nicons  : int
    <for each icon>
       nimages : int
       <for each image>
         min_dpi : int
         data_length: int
         data:  bytes
 ***********************************************************************/





static void gui_decode_icon(gui_icon_t *icon)
{
  int i,j;
  unsigned W, H;
  int length;
  unsigned char *byte_data;
  gui_iconimage_t *img;
  int dpi;

  dpi= getdpi();
  img= icon->images+0;
  for(i=1;i<icon->nimages;i++)
   if (dpi>icon->images[i].min_dpi)
     img= icon->images+i;
  //
  // we know the following won't fail because we did the same thing
  // while compiling the module and got no errors.
  //
  lodepng_decode32(&byte_data, &W, &H, img->data, img->data_length);
  icon->w= W;
  icon->h= H;
  icon->data= malloc(sizeof(icon->data[0])*W*H);
  length= W*H;
  for(i=0,j=0;i<length;i++,j+=4)
    icon->data[i]=   (byte_data[j+3]<<24) | (byte_data[j]<<16) 
                    | (byte_data[j+1]<<8) | (byte_data[j+2]) ;
  free(byte_data);
}

//
// icon_no is from 1 to gui_nicons
// 0 is used for no-icon situations
//
static gui_icon_t *gui_get_icon(int icon_no)
{
  gui_icon_t *icon;
  if (icon_no<1 || icon_no>gui_nicons) return NULL;
  icon_no--;
  icon= &gui_icons[icon_no];
  if (!icon->data) gui_decode_icon(icon);
  return icon;
}
