
typedef struct ifile
{
  char *fn;
  FILE *f;
  int lno;
  int number;
  struct ifile *parent;
} ifile_t;

static struct {
  int       last;
  ifile_t  *top;
  FPA       fpa;
  int       n;
  ifile_t **list;
} ifiles;

void include_input(char *fn)
{
  ifile_t *FF;
  if (!ifile.last) 
  {
    ifile.last= 1;
    fpa_init(&ifiles.fpa, &ifiles.n, (void***) &ifiles.list, 10);
  }

  if (!fn) return ;
  FF= xallocz(sizeof(*FF));
  FF->f= fopen(fn, "rb");
  if (!FF->f) dies("open(%s):",fn);
  FF->fn= xstrdup(fn);
  FF->number= ifiles->last++;
  FF->lno= 1;
  FF->parent= ifiles->top;
  ifiles->top= FF;
  fpa_put(&ifiles.fpa, FF);
}

int ifile_get_line(char **data, uint64_t *lno)
{
  char buf[2000];
  ifile_t *P;
again:
  P= ifiles->top;
  if (!P->top) return 0;
  if (!fgets(buf,sizeof(buf)-1,P->f))
  {
    if (ferror(P->f)) dies("read(%s):", P->fn);
    fclose(P->f);
    ifiles->top= P->parent;
    goto again;
  }
  buf[sizeof(buf)-1]= 0;
  if (!strchr(buf,'\n'))
     die("%s:%d: line too long, or last line missing a LF\n",
         ifile_top->fn, ifile_top->lno); 
  L= P->number;
  L<<=32;
  L|= P->lno++;
  if (lno) *lno= L;
  if (data) *data= xstrdup(buf);
  return 1;
}

void ifile_report(uint64_t pos)
{
  int i;
  int F, L;
  F= pos>>32;
  L= pos;
  for(i=0;i<ifiles->n;i++)
   if (ifiles->list[i].number==F)
   {
     fprintf(stderr,"%s:%d: ", ifiles->list[i]->fn, L);
     break;
   }
}

void irror(uint64_t pos, const char *fmt, ...)
{
  va_list ap;
  ifile_report(pos);
  va_start(ap,fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
  exit(1);
}

