#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <dirent.h>
#include <unistd.h>
#include <errno.h>
#include "../clay/components/134-miniz-defs.c"
#include "../clay/components/135-miniz.c"

#define MAXFIL 1000
#define N_DIGITS 3

static size_t smin2(size_t a,size_t b) { return a<b?a:b; }

#define IN_BUFFER_SIZE  2048
#define OUT_BUFFER_SIZE 4096

static void write_bytes(unsigned char *B, int L)
{
  int i,W;
  unsigned char K;
  static int x;
  if (!B) { x= 0; return ; }
  for(i=0;i<L;i++)
  {
    K= B[i];
    W= K/100 ? 4 : (K/10?3:2);
    if (W+x>76) { printf("\n"); x= 0; }
    printf("%d,",K);
    x+= W;
  }
}


static int output_file(char *fn)
{
  FILE *inf;
  int R;
  int flush;
  size_t inp_size;
  size_t olen;
  size_t done;
  mz_stream MZ;
  size_t inp_done;
  int nr;
  unsigned char inbuffer[IN_BUFFER_SIZE];
  unsigned char outbuffer[OUT_BUFFER_SIZE];

  fprintf(stderr,"compressing %s\n", fn);

  inf= fopen(fn,"rb");
  fseek(inf, 0, SEEK_END);
  inp_size= ftell(inf);
  fseek(inf, 0, SEEK_SET);

  inp_done= 0;
  printf(" (unsigned char[]) {\n");
  memset(&MZ, 0, sizeof(MZ));
  MZ.next_in= inbuffer;
  MZ.avail_in= 0;
  MZ.next_out= outbuffer;
  MZ.avail_out= OUT_BUFFER_SIZE;

  mz_deflateInit(&MZ, 9);
  done= 0;
  flush= MZ_NO_FLUSH;

  while(1)
  {
    if (MZ.avail_in==0)
    {
      nr= smin2(IN_BUFFER_SIZE, inp_size-inp_done);
      R= fread(inbuffer, 1, nr,inf);
      if (R!=nr)
      {
        fprintf(stderr,"can't read file\n");
        return 1;
      }
      MZ.next_in= inbuffer;
      MZ.avail_in= nr;
      inp_done+= nr;
    }

    R= mz_deflate(&MZ, inp_done==inp_size? MZ_FINISH:MZ_NO_FLUSH );
    if ((R==MZ_STREAM_END) || (!MZ.avail_out))
    {
      olen= OUT_BUFFER_SIZE - MZ.avail_out;
      if (olen) write_bytes(outbuffer, olen);
      done+= olen;
      MZ.next_out= outbuffer;
      MZ.avail_out= OUT_BUFFER_SIZE;
    }

    if (R==MZ_STREAM_END) break; 

    if (R!=MZ_OK) { fprintf(stderr,"deflate failed with code %d\n",R);
              fprintf(stderr,"read bytes so far: %d\n", (int) inp_done);
                      return 1; }
  }
  printf("}, %d \n", (int) done);
  fclose(inf);
  mz_deflateEnd(&MZ);
  fprintf(stderr,"done %s, compressed %d bytes\n", fn, (int) inp_done);
  return 0;
}

char *pname;
static char *files[MAXFIL];
static int n_files;

void read_files()
{
  DIR *dir;
  int i;
  struct dirent *de;
  char *name;
  dir= opendir(".");
  n_files= 0;
  if (!dir) 
    exit(fprintf(stderr,"can't open $WD: %s\n", strerror(errno)));
  while((de=readdir(dir)))
  {
    if (n_files>=MAXFIL) 
      exit(fprintf(stderr,"%s: too many input files, "
                          "recompile with larger MAXFIL\n", pname));
    name= de->d_name;
    for(i=0;i<N_DIGITS;i++) if (!isdigit(name[i])) break;
    if (i!=N_DIGITS) continue;
    if (name[i]!='-') continue;
    files[n_files]= strdup(name);
    n_files++;
  }
  closedir(dir);
  if (!n_files)
   exit(fprintf(stderr,"%s: no input files found\n", pname));
}

static void sort_files()
{
  int i,j;
  char *T;
  for(i=0;i<n_files-1;i++)
  {
    for(j=i+1;j<n_files;j++)
      if (strcmp(files[i],files[j])>0)
      {
        T= files[i];
        files[i]= files[j];
        files[j]= T;
      }
  }
}

static void mkfile(char *fn)
{
  printf(" { \"%s\", \n",fn);
  if (output_file(fn)) exit(1);
  printf(" },\n");
}

// the only argument is the path to the components directory
int main(int argc,char **argv)
{
  int i;
  char *aname, *fname;
  if (argc!=2)
   exit(fprintf(stderr,"usage: %s components-dir\n",argv[0]));

  if (chdir(argv[1]))
   exit(fprintf(stderr,"%s: chdir(%s): %s\n",argv[0],argv[1],strerror(errno)));

#if 0
  read_files();
  sort_files();

  for(i=0;i<n_files;i++)
  {
    printf(" { \"%s\", \n",files[i]+N_DIGITS+1);
    if (output_file(files[i])) return 1;
    printf(" },\n");
  }
#endif
  mkfile("libgfx.h");
  mkfile("libgfx.c");
  mkfile("lodepng.h");
  mkfile("stb_truetype.h");
  return 0;
}
