#include <stdio.h>

typedef struct { int d0, d1, target; } malink_t;
typedef struct { int type, target; char *text; } matim_t;
typedef struct { int nitem, maxw; matim_t *item; } matoc_t; 

/***
    The first char of each line string indicates its type.

    "!Chapter"
    "+Section"
    "=Paragraph"
 ***/
#define MA_CHAP   '!'
#define MA_SECT   '+'
#define MA_NORM   '='

typedef struct
{
  int nline;     char **line;
  int nlink;     malink_t *link;
  int ntoc;      matoc_t *toc;
} man_t;

#define MA_HEIGHT 24   
#define MA_WIDTH  80
#define MA_UPCTX  3     // number of lines to show before a target line

#define MA_NORM   0
#define MA_BOLD   1
#define MA_RVRS   2

typedef struct { int att; unsigned char data; } mac_t;

typedef struct
{
  int l0;  // the top-most line number
  mac_t **matrix;
  int u0, u1;
  int cx, cy; // cursor pos
  int w, h;
  const man_t *input;
} mao_t;

int imin2(int a,int b) {return a<b?a:b;}
int imax2(int a,int b) {return a>b?a:b;}

void tjump(int line, int column) { printf("\033[%d;%df", line, column); }
void tattr(int att) 
{
  switch(attr) {
  case MA_BOLD: printf("\0331m"); break;
  case MA_NORM: printf("\0330m"); break;
  case MA_RVRS: printf("\033?5h"); break;
  } 
}
void tscrollup(int n) { printf("\033[%dS", n); }
void tscrolldown(int n) { printf("\033[%dT",n); }


mao_t* manew(int w, int h)
{
  mao_t *M;
  M= malloc(sizeof(*M));
  M->w= w <= 0 ? MA_WIDTH : w;
  M->h= h <= 0 ? MA_HEIGHT: h;
  M->cx= M->cy= 0;
  M->matrix= malloc(sizeof(*matrix)*h);
  for(i=0;i<h;i++) M->matrix[i]= malloc(sizeof(*M->matrix[i])*w);
  return M;
}

void maupl(mao_t *M, int n)
{
  if (!n) return ;
  M->u0= imin2(M->u0, M->cy*M->w+M->cx);
  M->cx+= n;
  M->u1= imax2(M->u1, M->cy*M->w+M->cx-1);
}

void maprt(mao_t *M, char *s)
{
  int i;
  for(i=0;s[i];i++)
  {
    if (M->cx+i>=M->w) break;
    M->matrix[M->cy][M->cx+i].data= s[i];
    M->matrix[M->cy][M->cx+i].attr= MA_NORM;
  }
  maupl(M, i);
}

void machr(mao_t *M, char s)
{
  if (M->cx>=M->w) return ;
  M->matrix[M->cy][M->cx].data= s;
  maupl(M, 1);
}

void macspc(mao_t *M,int n)
{
  int i;
  n= imin2(n, M->w-M->cx);
  for(i=0;i<n;i++)
  {
    M->matrix[M->cy][M->cx+i].data= ' ';
    M->matrix[M->cy][M->cx+i].attr= MA_NORM;
  }
  maupl(M, n);
}

void maclrln(mao_t *M)
{
  maspc(M, M->w-M->cx);
}

void majmp(mao_t *M, int y, int x)
{
  M->cy= y;
  M->cx= x;
}

void mashow(mao_t *M)
{
  int y, x, y0, x0, x1, y1, sx, sy, a, n;
  char buf[200]; // maximum width of a line. should be 480 etc for UTF-8

  M->u0= imax2(M->u0, 0);
  M->u1= imin2(M->u1, M->w*M->h);

  y0= M->u0/M->w; x0= M->u0%M->w;
  y1= M->u1/M->w; x1= M->u1%M->w;

  a= -1;
  for(y=y0;y<=y1;y++)
  {
    sx= y==y0 ? x0 : 0;
    ex= y==y1 ? x1 : M->w-1;
    n= 0;
    for(x=sx;x<=ex;x++)
    {
      if (M->matrix[y][x].attr==a)
      {
        buf[n++]= M->matrix[y][x].data;
        continue;
      }
      if (n)
      {
        buf[n]= 0;
        tgoto(y,sx);
        tattr(a);
        printf("%s", buf);
      } 
      a= M->matrix[y][x].attr; 
      sx+= n;
      x--;
      n= 0;
    }
    if (n)
    {
      buf[n]= 0;
      tgoto(y,sx);
      tattr(a);
      printf("%s", buf);
    } 
  }
}

void maattr(mao_t *M, int p0, int p1, int a)
{
  int y, x, y0, x0, x1, y1, sx, sy;

  y0= p0/M->w; x0= p0%M->w;
  y1= p1/M->w; x1= p1%M->w;

  for(y=y0;y<=y1;y++)
  {
    sx= y==y0 ? x0 : 0;
    ex= y==y1 ? x1 : M->w-1;
    for(x=sx;x<=ex;x++) M->matrix[y][x].attr= a;
  }

  M->u0= imin2(M->u0, p0);
  M->u1= imin2(M->u1, p1);
}

int madisp(mao_t *M, int sy, int n)
{
  int n, l;
  const man_t *I;
  int l;
  malink_t *L;

  if (n<=0) return ;
  I= M->input;
  l= M->l0 + sy;
  n= imin2(M->input->nline-ls, M->h-1-sy);
  if (!n) return ;
  L= mafind(I, sy*M->w+0);
  for(d=0;d<n;d++)
  {
    majmp(M, sy+d, 0);
    maprt(M, I->line[l+d]+1);
    maclrln(M);
    switch(I->line[l+d][0])
    {
    case MA_CHAP:
    case MA_SECT:
      maattr(M, M->cy*M->w+0, M->cy*M->w+M->w-1, MA_BOLD);
      continue;
    }
    if (!L) continue;
again:
    p0= (sy+d)*M->w;
    p1= p0+M->w-1;
    switch(macommon(L, -M->l0*M->w, &p0, &p1))
    {
    case 0: maattr(M, p0, p1, MA_BOLD); break;
    case -1: L++; goto again;
    case 1: break;
    case 2: L= NULL; break;
    }
  }
}

void mastatus(mao_t *M)
{
  majmp(M, M->h-1, 0);
  maprt(M, "This is the way of life");
  maattr(M, (M->h-1)*M->w, (M->h-1)*M->w+M->w-1, MA_RVRS);
}

void mainit(mao_t *M, const man_t *I)
{
  M->input= I;
  M->l0= 0;
  madisp(M, 0, M->h-1);
  mastatus(M);
}


int main()
{
  mao_t *M;
  M= manew(0,0);
  mainit(M, myman);
}
